"""Test sparse histogram behavior."""

import numpy as np
import numpy.testing as npt
import pytest
from numpy import frombuffer

import hats.pixel_math.healpix_shim as hp
from hats.pixel_math.sparse_histogram import HistogramAggregator, SparseHistogram


def test_make_empty():
    """Tests the initialization of an empty histogram at the specified order"""
    histogram = SparseHistogram([], [], 5)
    expected_hist = np.zeros(hp.order2npix(5))
    npt.assert_array_equal(expected_hist, histogram.to_array())


def test_read_write_round_trip(tmp_path):
    """Test that we can read what we write into a histogram file."""
    histogram = SparseHistogram([11], [131], 0)

    # Write as a sparse array
    file_name = tmp_path / "round_trip_sparse.npz"
    histogram.to_file(file_name)
    read_histogram = SparseHistogram.from_file(file_name)
    npt.assert_array_equal(read_histogram.to_array(), histogram.to_array())

    # Write as a dense 1-d numpy array
    file_name = tmp_path / "round_trip_dense.npz"
    histogram.to_dense_file(file_name)
    with open(file_name, "rb") as file_handle:
        read_histogram = frombuffer(file_handle.read(), dtype=np.int64)
    npt.assert_array_equal(read_histogram, histogram.to_array())


def test_add_same_order():
    """Test that we can add two histograms created from the same order, and get
    the expected results."""
    partial_histogram_left = SparseHistogram([11], [131], 0)

    partial_histogram_right = SparseHistogram([10, 11], [4, 15], 0)

    total_histogram = HistogramAggregator(0)
    total_histogram.add(partial_histogram_left)
    total_histogram.add(partial_histogram_right)

    expected = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 146]
    npt.assert_array_equal(total_histogram.full_histogram, expected)

    sparse = total_histogram.to_sparse()
    npt.assert_array_equal(sparse.indexes, [10, 11])
    npt.assert_array_equal(sparse.counts, [4, 146])
    npt.assert_array_equal(sparse.order, 0)


def test_add_different_order():
    """Test that we can NOT add histograms of different healpix orders."""
    partial_histogram_left = SparseHistogram([11], [131], 0)
    partial_histogram_right = SparseHistogram([10, 11], [4, 15], 1)

    total_histogram = HistogramAggregator(0)
    total_histogram.add(partial_histogram_left)
    with pytest.raises(ValueError, match="partials have incompatible sizes"):
        total_histogram.add(partial_histogram_right)


def test_add_different_type():
    """Test that we can NOT add histograms of different healpix orders."""
    partial_histogram_left = SparseHistogram([11], [131], 0)

    total_histogram = HistogramAggregator(0)
    total_histogram.add(partial_histogram_left)
    with pytest.raises(ValueError, match="addends should be SparseHistogram"):
        total_histogram.add(5)

    with pytest.raises(ValueError, match="addends should be SparseHistogram"):
        total_histogram.add(([1, 2, 3, 4, 5], [1, 2, 3, 4, 5], 0))



"""Tests of file IO (reads and writes)"""

import shutil

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from hats.io import file_io, paths
from hats.io.parquet_metadata import aggregate_column_statistics, per_pixel_statistics, write_parquet_metadata
from hats.pixel_math.healpix_pixel import HealpixPixel


def test_write_parquet_metadata(tmp_path, small_sky_dir, small_sky_schema, check_parquet_schema):
    """Copy existing catalog and create new metadata files for it"""
    catalog_base_dir = tmp_path / "catalog"
    shutil.copytree(
        small_sky_dir,
        catalog_base_dir,
    )
    total_rows = write_parquet_metadata(catalog_base_dir)
    assert total_rows == 131
    check_parquet_schema(catalog_base_dir / "dataset" / "_metadata", small_sky_schema)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        catalog_base_dir / "dataset" / "_common_metadata",
        small_sky_schema,
        0,
    )
    ## Re-write - should still have the same properties.
    total_rows = write_parquet_metadata(catalog_base_dir)
    assert total_rows == 131
    check_parquet_schema(catalog_base_dir / "dataset" / "_metadata", small_sky_schema)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        catalog_base_dir / "dataset" / "_common_metadata",
        small_sky_schema,
        0,
    )


def test_write_parquet_metadata_order1(
    tmp_path, small_sky_order1_dir, small_sky_schema, check_parquet_schema
):
    """Copy existing catalog and create new metadata files for it,
    using a catalog with multiple files."""
    temp_path = tmp_path / "catalog"
    shutil.copytree(
        small_sky_order1_dir,
        temp_path,
    )

    total_rows = write_parquet_metadata(temp_path)
    assert total_rows == 131
    ## 4 row groups for 4 partitioned parquet files
    check_parquet_schema(
        temp_path / "dataset" / "_metadata",
        small_sky_schema,
        4,
    )
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        temp_path / "dataset" / "_common_metadata",
        small_sky_schema,
        0,
    )


def test_write_parquet_metadata_sorted(
    tmp_path, small_sky_order1_dir, small_sky_schema, check_parquet_schema
):
    """Copy existing catalog and create new metadata files for it,
    using a catalog with multiple files."""
    temp_path = tmp_path / "catalog"
    shutil.copytree(
        small_sky_order1_dir,
        temp_path,
    )

    total_rows = write_parquet_metadata(temp_path)
    assert total_rows == 131
    ## 4 row groups for 4 partitioned parquet files
    check_parquet_schema(
        temp_path / "dataset" / "_metadata",
        small_sky_schema,
        4,
    )
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        temp_path / "dataset" / "_common_metadata",
        small_sky_schema,
        0,
    )


def test_write_index_parquet_metadata(tmp_path, check_parquet_schema):
    """Create an index-like catalog, and test metadata creation."""
    temp_path = tmp_path / "index"

    index_parquet_path = temp_path / "dataset" / "Parts=0" / "part_000_of_001.parquet"
    file_io.make_directory(temp_path / "dataset" / "Parts=0")
    basic_index = pd.DataFrame({"_healpix_29": [4000, 4001], "ps1_objid": [700, 800]})
    file_io.write_dataframe_to_parquet(basic_index, index_parquet_path)

    index_catalog_parquet_metadata = pa.schema(
        [
            pa.field("_healpix_29", pa.int64()),
            pa.field("ps1_objid", pa.int64()),
        ]
    )

    total_rows = write_parquet_metadata(temp_path, order_by_healpix=False)
    assert total_rows == 2

    check_parquet_schema(tmp_path / "index" / "dataset" / "_metadata", index_catalog_parquet_metadata)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        tmp_path / "index" / "dataset" / "_common_metadata",
        index_catalog_parquet_metadata,
        0,
    )


def test_aggregate_column_statistics(small_sky_order1_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)

    result_frame = aggregate_column_statistics(partition_info_file)
    assert len(result_frame) == 5

    result_frame = aggregate_column_statistics(partition_info_file, exclude_hats_columns=False)
    assert len(result_frame) == 9

    result_frame = aggregate_column_statistics(partition_info_file, include_columns=["ra", "dec"])
    assert len(result_frame) == 2

    result_frame = aggregate_column_statistics(partition_info_file, include_columns=["does", "not", "exist"])
    assert len(result_frame) == 0


def assert_column_stat_as_floats(
    result_frame, column_name, min_value=None, max_value=None, null_count=0, row_count=None
):
    assert column_name in result_frame.index
    data_stats = result_frame.loc[column_name]
    assert float(data_stats["min_value"]) >= min_value
    assert float(data_stats["max_value"]) <= max_value
    assert int(data_stats["null_count"]) == null_count
    assert int(data_stats["row_count"]) == row_count


def test_aggregate_column_statistics_with_pixel(small_sky_order1_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)

    result_frame = aggregate_column_statistics(partition_info_file)
    assert len(result_frame) == 5
    assert_column_stat_as_floats(result_frame, "dec", min_value=-69.5, max_value=-25.5, row_count=131)

    result_frame = aggregate_column_statistics(partition_info_file, include_pixels=[HealpixPixel(1, 45)])
    assert len(result_frame) == 5
    assert_column_stat_as_floats(result_frame, "dec", min_value=-60.5, max_value=-25.5, row_count=29)

    result_frame = aggregate_column_statistics(partition_info_file, include_pixels=[HealpixPixel(1, 47)])
    assert len(result_frame) == 5
    assert_column_stat_as_floats(result_frame, "dec", min_value=-36.5, max_value=-25.5, row_count=18)

    result_frame = aggregate_column_statistics(
        partition_info_file, include_pixels=[HealpixPixel(1, 45), HealpixPixel(1, 47)]
    )
    assert len(result_frame) == 5
    assert_column_stat_as_floats(result_frame, "dec", min_value=-60.5, max_value=-25.5, row_count=47)

    result_frame = aggregate_column_statistics(partition_info_file, include_pixels=[HealpixPixel(1, 4)])
    assert len(result_frame) == 0


def test_aggregate_column_statistics_with_nested(small_sky_nested_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_nested_dir)

    ## Will have 13 returned columns (5 object and 8 light curve)
    ## Since object_dec is copied from object.dec, the min/max are the same,
    ## but there are MANY more rows of light curve dec values.
    result_frame = aggregate_column_statistics(partition_info_file)
    assert len(result_frame) == 13
    assert_column_stat_as_floats(result_frame, "dec", min_value=-69.5, max_value=-25.5, row_count=131)
    assert_column_stat_as_floats(
        result_frame, "lc.object_dec", min_value=-69.5, max_value=-25.5, row_count=16135
    )

    ## Only peeking at a single pixel, we should see the same dec min/max as
    ## we see above for the flat object table.
    result_frame = aggregate_column_statistics(partition_info_file, include_pixels=[HealpixPixel(1, 47)])
    assert len(result_frame) == 13
    assert_column_stat_as_floats(result_frame, "dec", min_value=-36.5, max_value=-25.5, row_count=18)
    assert_column_stat_as_floats(
        result_frame, "lc.source_id", min_value=70008, max_value=87148, row_count=2358
    )
    assert_column_stat_as_floats(result_frame, "lc.mag", min_value=15, max_value=21, row_count=2358)

    ## Test that we can request light curve columns, using the shorter name
    ## e.g. full path in the file is "lc.source_id.list.element"
    result_frame = aggregate_column_statistics(
        partition_info_file, include_columns=["ra", "dec", "lc.source_ra", "lc.source_dec", "lc.mag"]
    )
    assert len(result_frame) == 5
    assert_column_stat_as_floats(result_frame, "dec", min_value=-69.5, max_value=-25.5, row_count=131)
    assert_column_stat_as_floats(result_frame, "lc.mag", min_value=15, max_value=21, row_count=16135)


def test_aggregate_column_statistics_with_nulls(tmp_path):
    file_io.make_directory(tmp_path / "dataset")

    metadata_filename = tmp_path / "dataset" / "dataframe_01.parquet"
    table_with_schema = pa.Table.from_arrays([[-1.0], [1.0]], names=["data", "Npix"])
    pq.write_table(table_with_schema, metadata_filename)

    icky_table = pa.Table.from_arrays([[2.0, None], [None, 6.0]], schema=table_with_schema.schema)
    metadata_filename = tmp_path / "dataset" / "dataframe_02.parquet"
    pq.write_table(icky_table, metadata_filename)

    icky_table = pa.Table.from_arrays([[None], [None]], schema=table_with_schema.schema)
    metadata_filename = tmp_path / "dataset" / "dataframe_00.parquet"
    pq.write_table(icky_table, metadata_filename)

    icky_table = pa.Table.from_arrays([[None, None], [None, None]], schema=table_with_schema.schema)
    metadata_filename = tmp_path / "dataset" / "dataframe_03.parquet"
    pq.write_table(icky_table, metadata_filename)

    assert write_parquet_metadata(tmp_path, order_by_healpix=False) == 6

    result_frame = aggregate_column_statistics(tmp_path / "dataset" / "_metadata", exclude_hats_columns=False)
    assert len(result_frame) == 2

    assert_column_stat_as_floats(result_frame, "data", min_value=-1, max_value=2, null_count=4, row_count=6)
    assert_column_stat_as_floats(result_frame, "Npix", min_value=1, max_value=6, null_count=4, row_count=6)


def test_per_pixel_statistics(small_sky_order1_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)

    result_frame = per_pixel_statistics(partition_info_file)
    # 20 = 5 columns * 4 stats per-column
    assert result_frame.shape == (4, 20)

    result_frame = per_pixel_statistics(partition_info_file, exclude_hats_columns=False)
    # 36 = 9 columns * 4 stats per-column
    assert result_frame.shape == (4, 36)

    result_frame = per_pixel_statistics(partition_info_file, include_columns=["ra", "dec"])
    # 8 = 2 columns * 4 stats per-column
    assert result_frame.shape == (4, 8)

    result_frame = per_pixel_statistics(partition_info_file, include_columns=["does", "not", "exist"])
    assert len(result_frame) == 0


def test_per_pixel_statistics_multi_index(small_sky_order1_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)

    result_frame = per_pixel_statistics(partition_info_file, multi_index=True)
    # 20 = 5 columns * 4 pixels
    assert result_frame.shape == (20, 4)

    result_frame = per_pixel_statistics(partition_info_file, exclude_hats_columns=False, multi_index=True)
    # 36 = 9 columns * 4 stats per-column
    assert result_frame.shape == (36, 4)


def test_per_pixel_statistics_include_stats(small_sky_order1_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)

    result_frame = per_pixel_statistics(partition_info_file, include_stats=["row_count"])
    # 5 = 5 columns * 1 stat per column
    assert result_frame.shape == (4, 5)

    result_frame = per_pixel_statistics(
        partition_info_file, include_stats=["row_count"], include_columns=["id"]
    )
    # 1 = 1 columns * 1 stat per column
    assert result_frame.shape == (4, 1)

    result_frame = per_pixel_statistics(
        partition_info_file, include_stats=["row_count"], include_columns=["id"], multi_index=True
    )
    # 1 = 1 columns * 1 stat per column
    assert result_frame.shape == (4, 1)



from __future__ import annotations

import warnings
from pathlib import Path

import numpy as np
import pyarrow as pa
from mocpy import MOC
from upath import UPath

import hats.pixel_math.healpix_shim as hp
from hats.catalog import AssociationCatalog, Catalog, CatalogType, Dataset, MapCatalog, MarginCatalog
from hats.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hats.catalog.dataset.table_properties import TableProperties
from hats.catalog.index.index_catalog import IndexCatalog
from hats.catalog.partition_info import PartitionInfo
from hats.io import file_io, paths
from hats.io.file_io import read_parquet_metadata

DATASET_TYPE_TO_CLASS = {
    CatalogType.OBJECT: Catalog,
    CatalogType.SOURCE: Catalog,
    CatalogType.ASSOCIATION: AssociationCatalog,
    CatalogType.INDEX: IndexCatalog,
    CatalogType.MARGIN: MarginCatalog,
    CatalogType.MAP: MapCatalog,
}


def read_hats(catalog_path: str | Path | UPath) -> Dataset:
    """Reads a HATS Catalog from a HATS directory

    Args:
        catalog_path (str): path to the root directory of the catalog

    Returns:
        The initialized catalog object

    Examples:
        To read a catalog from a public S3 bucket, call it as follows::

            from upath import UPath
            catalog = hats.read_hats(UPath(..., anon=True))
    """
    catalog_path = file_io.get_upath(catalog_path)
    try:
        properties = TableProperties.read_from_dir(catalog_path)
        dataset_type = properties.catalog_type
        if dataset_type not in DATASET_TYPE_TO_CLASS:
            raise NotImplementedError(f"Cannot load catalog of type {dataset_type}")

        loader = DATASET_TYPE_TO_CLASS[dataset_type]
        kwargs = {
            "catalog_path": catalog_path,
            "catalog_info": properties,
            "schema": _read_schema_from_metadata(catalog_path),
        }
        if _is_healpix_dataset(dataset_type):
            kwargs["pixels"] = PartitionInfo.read_from_dir(catalog_path)
            kwargs["moc"] = _read_moc_from_point_map(catalog_path)
        if dataset_type == CatalogType.ASSOCIATION:
            kwargs["join_pixels"] = PartitionJoinInfo.read_from_dir(catalog_path)
        return loader(**kwargs)
    except Exception as exception:  # pylint: disable=broad-except
        raise FileNotFoundError(f"Failed to read HATS at location {catalog_path}") from exception


def _is_healpix_dataset(dataset_type):
    return dataset_type in (
        CatalogType.OBJECT,
        CatalogType.SOURCE,
        CatalogType.ASSOCIATION,
        CatalogType.MARGIN,
        CatalogType.MAP,
    )


def _read_moc_from_point_map(catalog_base_dir: str | Path | UPath) -> MOC | None:
    """Reads a MOC object from the `point_map.fits` file if it exists in the catalog directory"""
    point_map_path = paths.get_point_map_file_pointer(catalog_base_dir)
    if not file_io.does_file_or_directory_exist(point_map_path):
        return None
    fits_image = file_io.read_fits_image(point_map_path)
    order = hp.npix2order(len(fits_image))
    boolean_skymap = fits_image.astype(bool)
    ipix = np.where(boolean_skymap)[0]
    orders = np.full(ipix.shape, order)
    return MOC.from_healpix_cells(ipix, orders, order)


def _read_schema_from_metadata(catalog_base_dir: str | Path | UPath) -> pa.Schema | None:
    """Reads the schema information stored in the _common_metadata or _metadata files."""
    common_metadata_file = paths.get_common_metadata_pointer(catalog_base_dir)
    common_metadata_exists = file_io.does_file_or_directory_exist(common_metadata_file)
    metadata_file = paths.get_parquet_metadata_pointer(catalog_base_dir)
    metadata_exists = file_io.does_file_or_directory_exist(metadata_file)
    if not (common_metadata_exists or metadata_exists):
        warnings.warn(
            "_common_metadata or _metadata files not found for this catalog."
            "The arrow schema will not be set."
        )
        return None
    schema_file = common_metadata_file if common_metadata_exists else metadata_file
    metadata = read_parquet_metadata(schema_file)
    return metadata.schema.to_arrow_schema()



from unittest.mock import MagicMock

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pytest
from astropy.coordinates import Angle, SkyCoord
from astropy.visualization.wcsaxes.frame import EllipticalFrame, RectangularFrame
from matplotlib import colors
from matplotlib.colors import LogNorm, Normalize
from matplotlib.path import Path
from matplotlib.pyplot import get_cmap
from mocpy import MOC, WCS
from mocpy.moc.plot import fill
from mocpy.moc.plot.culling_backfacing_cells import from_moc
from mocpy.moc.plot.utils import build_plotting_moc

from hats import read_hats
from hats.inspection import plot_density, plot_pixels
from hats.inspection.visualize_catalog import (
    compute_healpix_vertices,
    cull_from_pixel_map,
    cull_to_fov,
    plot_healpix_map,
    plot_moc,
)

# pylint: disable=no-member


DEFAULT_CMAP_NAME = "viridis"
DEFAULT_FOV = (320 * u.deg, 160 * u.deg)
DEFAULT_CENTER = SkyCoord(0, 0, unit="deg", frame="icrs")
DEFAULT_COORDSYS = "icrs"
DEFAULT_ROTATION = Angle(0, u.degree)
DEFAULT_PROJECTION = "MOL"


@pytest.fixture(autouse=True)
def reset_matplotlib():
    yield
    plt.close("all")


def test_healpix_vertices():
    depth = 3
    ipix = np.array([10, 11])
    fig = plt.figure()
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    paths, codes = compute_healpix_vertices(depth, ipix, wcs)
    mocpy_paths, mocpy_codes = fill.compute_healpix_vertices(depth, ipix, wcs)
    np.testing.assert_array_equal(paths, mocpy_paths)
    np.testing.assert_array_equal(codes, mocpy_codes)


def test_healpix_vertices_step():
    depth = 1
    ipix = np.array([10, 11])
    fig = plt.figure()
    step = 4
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    paths, codes = compute_healpix_vertices(depth, ipix, wcs, step=step)
    # checks the codes match the expected matplotlib path codes
    np.testing.assert_array_equal(
        codes, np.tile(np.array([Path.MOVETO] + [Path.LINETO] * (step * 4 - 1) + [Path.CLOSEPOLY]), len(ipix))
    )
    mocpy_paths, _ = fill.compute_healpix_vertices(depth, ipix, wcs)
    # mocpy only generates path points at the healpix corner vertices. So we subsample our generated vertices
    # to check the corners match the expected mocpy generated ones
    first_path_vertex_indices = np.array([0, step, 2 * step, 3 * step, 4 * step])
    start_path_index = np.array(([0] * 5) + ([first_path_vertex_indices[-1] + 1] * 5))
    vertex_indices = start_path_index + np.tile(first_path_vertex_indices, len(ipix))
    np.testing.assert_array_equal(paths[vertex_indices], mocpy_paths)


def test_plot_healpix_pixels():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth)
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    assert col.get_cmap() == get_cmap(DEFAULT_CMAP_NAME)
    assert isinstance(col.norm, Normalize)
    assert col.norm.vmin == min(pix_map)
    assert col.norm.vmax == max(pix_map)
    assert col.colorbar is not None
    assert col.colorbar.cmap == get_cmap(DEFAULT_CMAP_NAME)
    assert col.colorbar.norm == col.norm
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    for path, ipix in zip(paths, ipix):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)
    assert ax.frame_class == EllipticalFrame


def test_plot_healpix_pixels_different_order():
    order = 6
    length = 1000
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth)
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w

    all_verts, all_codes = compute_healpix_vertices(order, ipix, wcs)
    for i, (path, ipix) in enumerate(zip(paths, ipix)):
        verts, codes = all_verts[i * 5 : (i + 1) * 5], all_codes[i * 5 : (i + 1) * 5]
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)


def test_order_0_pixel_plots_with_step():
    map_value = 0.5
    order_0_pix = 4
    ipix = np.array([order_0_pix])
    pix_map = np.array([map_value])
    depth = np.array([0])
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth)
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    length = 1
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    all_verts, all_codes = compute_healpix_vertices(0, ipix, wcs, step=2**3)
    # assert the number of vertices == number of pixels * 4 sides per pixel * steps per side + 1 for the
    # close polygon
    assert len(all_verts) == len(ipix) * 4 * (2**3) + 1
    np.testing.assert_array_equal(paths[0].vertices, all_verts)
    np.testing.assert_array_equal(paths[0].codes, all_codes)
    np.testing.assert_array_equal(col.get_array(), np.full(length, fill_value=map_value))


def test_edge_pixels_split_to_order_7():
    map_value = 0.5
    order_0_pix = 2
    ipix = np.array([order_0_pix])
    pix_map = np.array([map_value])
    depth = np.array([0])
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth)
    assert len(ax.collections) > 0

    # Generate a dictionary of pixel indices that have sides that align with the meridian at ra = 180deg, the
    # right edge of the plot
    edge_pixels = {0: [order_0_pix]}
    for iter_ord in range(1, 8):
        edge_pixels[iter_ord] = [p * 4 + i for p in edge_pixels[iter_ord - 1] for i in (2, 3)]

    # Generate a dictionary of subpixels of the order 0 pixel that are not on the edge of the plot, i.e. the
    # pixels that should be in the culled plot
    non_edge_pixels = {}
    pixels_ord1 = np.arange(4 * order_0_pix, 4 * (order_0_pix + 1))
    non_edge_pixels[1] = pixels_ord1[~np.isin(pixels_ord1, edge_pixels[1])]
    for iter_ord in range(2, 8):
        pixels_ord = np.concatenate([np.arange(4 * pix, 4 * (pix + 1)) for pix in edge_pixels[iter_ord - 1]])
        non_edge_pixels[iter_ord] = pixels_ord[~np.isin(pixels_ord, edge_pixels[iter_ord])]
    col = ax.collections[0]
    paths = col.get_paths()

    # Check that the plotted paths match the non_edge_pixels generated
    length = sum(len(x) for x in non_edge_pixels.values())
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    ords = np.concatenate([np.full(len(x), fill_value=o) for o, x in non_edge_pixels.items()])
    pixels = np.concatenate([np.array(x) for _, x in non_edge_pixels.items()])
    for path, iter_ord, pix in zip(paths, ords, pixels):
        step = 1 if iter_ord >= 3 else 2 ** (3 - iter_ord)
        verts, codes = compute_healpix_vertices(iter_ord, np.array([pix]), wcs, step=step)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), np.full(length, fill_value=map_value))


def test_cull_from_pixel_map():
    order = 3
    ipix = np.arange(12 * 4**order)
    pix_map = np.arange(12 * 4**order)
    map_dict = {order: (ipix, pix_map)}
    fig = plt.figure(figsize=(10, 5))
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    culled_dict = cull_from_pixel_map(map_dict, wcs)
    mocpy_culled = from_moc({str(order): ipix}, wcs)
    for iter_ord, (pixels, m) in culled_dict.items():
        np.testing.assert_array_equal(pixels, mocpy_culled[str(iter_ord)])
        map_indices = pixels >> (2 * (iter_ord - order))
        np.testing.assert_array_equal(m, pix_map[map_indices])


def test_cull_to_fov():
    order = 4
    ipix = np.arange(12 * 4**order)
    pix_map = np.arange(12 * 4**order)
    map_dict = {order: (ipix, pix_map)}
    fig = plt.figure(figsize=(10, 5))
    wcs = WCS(
        fig,
        fov=(20 * u.deg, 10 * u.deg),
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    culled_dict = cull_to_fov(map_dict, wcs)
    moc = MOC.from_healpix_cells(ipix, np.full(ipix.shape, fill_value=order), max_depth=order)
    mocpy_culled = build_plotting_moc(moc, wcs)
    for iter_ord, (pixels, m) in culled_dict.items():
        for p in pixels:
            assert (
                len(
                    MOC.from_healpix_cells(np.array([p]), np.array([iter_ord]), max_depth=iter_ord)
                    .intersection(mocpy_culled)
                    .to_depth29_ranges
                )
                > 0
            )
        ord_ipix = np.arange(12 * 4**iter_ord)
        ord_non_pixels = ord_ipix[~np.isin(ord_ipix, pixels)]
        for p in ord_non_pixels:
            assert (
                len(
                    MOC.from_healpix_cells(np.array([p]), np.array([iter_ord]), max_depth=iter_ord)
                    .intersection(mocpy_culled)
                    .to_depth29_ranges
                )
                == 0
            )
        map_indices = pixels >> (2 * (iter_ord - order))
        np.testing.assert_array_equal(m, pix_map[map_indices])


def test_cull_to_fov_subsamples_high_order():
    order = 10
    ipix = np.arange(12 * 4**order)
    pix_map = np.arange(12 * 4**order)
    map_dict = {order: (ipix, pix_map)}
    fig = plt.figure(figsize=(10, 5))
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    with pytest.warns(match="smaller"):
        culled_dict = cull_to_fov(map_dict, wcs)
    # Get the WCS cdelt giving the deg.px^(-1) resolution.
    cdelt = wcs.wcs.cdelt
    # Convert in rad.px^(-1)
    cdelt = np.abs((2 * np.pi / 360) * cdelt[0])
    # Get the minimum depth such as the resolution of a cell is contained in 1px.
    depth_res = int(np.floor(np.log2(np.sqrt(np.pi / 3) / cdelt)))
    depth_res = max(depth_res, 0)
    assert depth_res < order

    for iter_ord, (pixels, m) in culled_dict.items():
        assert iter_ord == depth_res
        assert np.all(np.isin(ipix >> (2 * (order - depth_res)), pixels))
        map_indices = pixels << (2 * (order - depth_res))
        np.testing.assert_array_equal(m, pix_map[map_indices])


def test_cull_to_fov_subsamples_multiple_orders():
    depth = np.array([0, 5, 8, 10])
    ipix = np.array([10, 5, 4, 2])
    pix_map = np.array([1, 2, 3, 4])
    map_dict = {depth[i]: (ipix[[i]], pix_map[[i]]) for i in range(len(depth))}
    fig = plt.figure(figsize=(10, 5))
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    with pytest.warns(match="smaller"):
        culled_dict = cull_to_fov(map_dict, wcs)
    # Get the WCS cdelt giving the deg.px^(-1) resolution.
    cdelt = wcs.wcs.cdelt
    # Convert in rad.px^(-1)
    cdelt = np.abs((2 * np.pi / 360) * cdelt[0])
    # Get the minimum depth such as the resolution of a cell is contained in 1px.
    depth_res = int(np.floor(np.log2(np.sqrt(np.pi / 3) / cdelt)))
    depth_res = max(depth_res, 0)
    assert depth_res < max(depth)

    assert list(culled_dict.keys()) == [0, 5, depth_res]

    assert culled_dict[0] == (np.array([10]), np.array([1]))
    assert culled_dict[5] == (np.array([5]), np.array([2]))
    small_pixels_map = pix_map[2:]
    small_pixels_converted = ipix[2:] >> (2 * (depth[2:] - depth_res))
    small_pixels_argsort = np.argsort(small_pixels_converted)
    assert np.all(culled_dict[depth_res][0] == small_pixels_converted[small_pixels_argsort])
    assert np.all(culled_dict[depth_res][1] == small_pixels_map[small_pixels_argsort])


def test_plot_healpix_map():
    order = 1
    ipix = np.arange(12 * 4**order)
    pix_map = np.arange(12 * 4**order)
    fig, ax = plot_healpix_map(pix_map)
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    map_dict = {order: (ipix, pix_map)}
    culled_dict = cull_from_pixel_map(map_dict, wcs)
    all_vals = []
    start_i = 0
    for iter_ord, (pixels, pix_map) in culled_dict.items():
        step = 1 if iter_ord >= 3 else 2 ** (3 - iter_ord)
        vert_len = step * 4 + 1
        all_verts, all_codes = compute_healpix_vertices(iter_ord, pixels, wcs, step=step)
        for i, _ in enumerate(pixels):
            verts, codes = (
                all_verts[i * vert_len : (i + 1) * vert_len],
                all_codes[i * vert_len : (i + 1) * vert_len],
            )
            path = paths[start_i + i]
            np.testing.assert_array_equal(path.vertices, verts)
            np.testing.assert_array_equal(path.codes, codes)
        all_vals.append(pix_map)
        start_i += len(pixels)
    assert start_i == len(paths)
    np.testing.assert_array_equal(np.concatenate(all_vals), col.get_array())


def test_plot_wcs_params():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(
        pix_map,
        ipix=ipix,
        depth=depth,
        fov=(100 * u.deg, 50 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        projection="AIT",
    )
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=(100 * u.deg, 50 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection="AIT",
    ).w
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)
    assert ax.get_transform("icrs") is not None
    assert ax.frame_class == RectangularFrame


def test_plot_wcs_params_frame():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(
        pix_map,
        ipix=ipix,
        depth=depth,
        fov=(100 * u.deg, 50 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        projection="AIT",
        frame_class=EllipticalFrame,
    )
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=(100 * u.deg, 50 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection="AIT",
    ).w
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)
    assert ax.get_transform("icrs") is not None
    assert ax.frame_class == EllipticalFrame


def test_plot_fov_culling():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(
        pix_map,
        ipix=ipix,
        depth=depth,
        fov=(30 * u.deg, 20 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        projection="AIT",
    )
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    wcs = WCS(
        fig,
        fov=(30 * u.deg, 20 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection="AIT",
    ).w
    map_dict = {order: (ipix, pix_map)}
    culled_dict = cull_to_fov(map_dict, wcs)
    all_vals = []
    start_i = 0
    for iter_ord, (pixels, pix_map) in culled_dict.items():
        all_verts, all_codes = compute_healpix_vertices(iter_ord, pixels, wcs)
        for i, _ in enumerate(pixels):
            verts, codes = all_verts[i * 5 : (i + 1) * 5], all_codes[i * 5 : (i + 1) * 5]
            path = paths[start_i + i]
            np.testing.assert_array_equal(path.vertices, verts)
            np.testing.assert_array_equal(path.codes, codes)
        all_vals.append(pix_map)
        start_i += len(pixels)
    assert start_i == len(paths)
    np.testing.assert_array_equal(np.concatenate(all_vals), col.get_array())
    assert ax.get_transform("icrs") is not None
    assert ax.frame_class == RectangularFrame


def test_plot_wcs():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig = plt.figure(figsize=(10, 5))
    wcs = WCS(
        fig,
        fov=(100 * u.deg, 50 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection="AIT",
    ).w
    fig2, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth, fig=fig, wcs=wcs)
    assert fig2 is fig
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)
    assert ax.get_transform("icrs") is not None


def test_plot_wcs_and_ax():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig = plt.figure(figsize=(10, 5))
    wcs = WCS(
        fig,
        fov=(100 * u.deg, 50 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection="AIT",
    ).w
    ax = fig.add_subplot(1, 1, 1, projection=wcs, frame_class=EllipticalFrame)
    assert len(ax.collections) == 0
    fig2, ax2 = plot_healpix_map(pix_map, ipix=ipix, depth=depth, fig=fig, wcs=wcs, ax=ax)
    assert fig2 is fig
    assert ax2 is ax
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)
    assert ax.get_transform("icrs") is not None


def test_plot_ax_no_wcs():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig = plt.figure(figsize=(10, 5))
    wcs = WCS(
        fig,
        fov=(100 * u.deg, 50 * u.deg),
        center=SkyCoord(10, 10, unit="deg", frame="icrs"),
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection="AIT",
    ).w
    ax = fig.add_subplot(1, 1, 1, projection=wcs, frame_class=EllipticalFrame)
    assert len(ax.collections) == 0
    fig2, ax2 = plot_healpix_map(pix_map, ipix=ipix, depth=depth, fig=fig, ax=ax)
    assert fig2 is fig
    assert ax2 is ax
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)
    assert ax.get_transform("icrs") is not None


def test_plot_cmaps():
    order = 3
    length = 10
    cmap_name = "plasma"
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth, cmap=cmap_name)
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert col.get_cmap() == get_cmap(cmap_name)
    assert col.colorbar is not None
    assert col.colorbar.cmap == get_cmap(cmap_name)
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)

    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth, cmap=get_cmap(cmap_name))
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert col.get_cmap() == get_cmap(cmap_name)
    assert col.colorbar is not None
    assert col.colorbar.cmap == get_cmap(cmap_name)
    assert len(paths) == length
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)


def test_plot_norm():
    order = 3
    length = 10
    norm = LogNorm()
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth, norm=norm)
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert col.norm == norm
    assert col.colorbar is not None
    assert col.colorbar.norm == norm
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)


def test_plot_no_cbar():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth, cbar=False)
    assert len(ax.collections) > 0
    col = ax.collections[0]
    assert col.colorbar is None
    paths = col.get_paths()
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)


def test_plot_kwargs():
    order = 3
    length = 10
    label = "test"
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig, ax = plot_healpix_map(pix_map, ipix=ipix, depth=depth, label=label)
    assert len(ax.collections) > 0
    col = ax.collections[0]
    assert col.get_label() == label
    paths = col.get_paths()
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)


def test_plot_existing_fig():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig = plt.figure()
    assert len(fig.axes) == 0
    fig_ret, ax_ret = plot_healpix_map(pix_map, ipix=ipix, depth=depth)
    assert fig is fig_ret
    ax = fig.get_axes()[0]
    assert ax is ax_ret
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)


def test_plot_existing_wcsaxes():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig = plt.figure()
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    ax = fig.add_subplot(1, 1, 1, projection=wcs)
    assert len(fig.axes) == 1
    assert len(ax.collections) == 0
    fig_ret, ax_ret = plot_healpix_map(pix_map, ipix=ipix, depth=depth)
    assert fig is fig_ret
    assert ax is ax_ret
    assert len(ax.collections) > 0
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == length
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)


def test_plot_existing_wrong_axes():
    order = 3
    length = 10
    ipix = np.arange(length)
    pix_map = np.arange(length)
    depth = np.full(length, fill_value=order)
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    assert len(fig.axes) == 1
    assert len(ax.collections) == 0
    with pytest.warns(match="WCSAxes"):
        fig_ret, ax_ret = plot_healpix_map(pix_map, ipix=ipix, depth=depth)
    assert fig is not fig_ret
    assert ax is not ax_ret
    assert len(ax.collections) == 0
    assert len(ax_ret.collections) == 1
    col = ax_ret.collections[0]
    paths = col.get_paths()
    wcs = WCS(
        fig_ret,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    assert len(paths) == length
    for path, ipix in zip(paths, np.arange(len(pix_map))):
        verts, codes = compute_healpix_vertices(order, np.array([ipix]), wcs)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), pix_map)


def test_catalog_plot_density(small_sky_source_dir):
    """Test plotting pixel-density for on-disk catalog.
    Confirm plotting at lower order doesn't have a warning, and creates fewer plot paths."""
    small_sky_source_catalog = read_hats(small_sky_source_dir)
    with pytest.warns(match="smaller"):
        _, ax = plot_density(small_sky_source_catalog)
    order10_paths = ax.collections[0].get_paths()
    assert "Angular density of catalog small_sky_source" == ax.get_title()

    _, ax = plot_density(small_sky_source_catalog, order=3)
    order3_paths = ax.collections[-1].get_paths()
    assert "Angular density of catalog small_sky_source" == ax.get_title()

    assert len(order3_paths) < len(order10_paths)


def test_catalog_plot_density_errors(small_sky_source_dir):
    small_sky_source_catalog = read_hats(small_sky_source_dir)
    with pytest.raises(ValueError, match="plotting order"):
        plot_density(small_sky_source_catalog, order=13)

    with pytest.raises(ValueError, match="not convertible"):
        plot_density(small_sky_source_catalog, unit="arcmin")

    with pytest.raises(ValueError, match="catalog required"):
        plot_density(None)


def test_catalog_plot(small_sky_order1_catalog):
    fig, ax = plot_pixels(small_sky_order1_catalog)
    pixels = sorted(small_sky_order1_catalog.get_healpix_pixels())
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == len(pixels)
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    for p, path in zip(pixels, paths):
        step = 2 ** (3 - p.order)
        verts, codes = compute_healpix_vertices(p.order, np.array([p.pixel]), wcs, step=step)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    np.testing.assert_array_equal(col.get_array(), np.array([p.order for p in pixels]))
    assert ax.get_title() == f"Catalog pixel density map - {small_sky_order1_catalog.catalog_name}"


def test_catalog_plot_no_color_by_order(small_sky_order1_catalog):
    fc = "white"
    ec = "black"
    fig, ax = plot_pixels(small_sky_order1_catalog, color_by_order=False, facecolor=fc, edgecolor=ec)
    pixels = sorted(small_sky_order1_catalog.get_healpix_pixels())
    col = ax.collections[0]
    paths = col.get_paths()
    assert len(paths) == len(pixels)
    wcs = WCS(
        fig,
        fov=DEFAULT_FOV,
        center=DEFAULT_CENTER,
        coordsys=DEFAULT_COORDSYS,
        rotation=DEFAULT_ROTATION,
        projection=DEFAULT_PROJECTION,
    ).w
    for p, path in zip(pixels, paths):
        step = 2 ** (3 - p.order)
        verts, codes = compute_healpix_vertices(p.order, np.array([p.pixel]), wcs, step=step)
        np.testing.assert_array_equal(path.vertices, verts)
        np.testing.assert_array_equal(path.codes, codes)
    assert col.get_array() is None
    np.testing.assert_array_equal(col.get_facecolor()[0], colors.to_rgba(fc))
    np.testing.assert_array_equal(col.get_edgecolor()[0], colors.to_rgba(ec))
    assert ax.get_title() == f"Catalog pixel density map - {small_sky_order1_catalog.catalog_name}"


def test_plot_moc(small_sky_order1_catalog):
    small_sky_order1_catalog.moc.fill = MagicMock()
    _, ax = plot_moc(small_sky_order1_catalog.moc)
    small_sky_order1_catalog.moc.fill.assert_called_once()
    assert small_sky_order1_catalog.moc.fill.call_args[0][0] is ax
    wcs = ax.wcs
    assert small_sky_order1_catalog.moc.fill.call_args[0][1] is wcs


def test_plot_moc_catalog(small_sky_order1_catalog):
    small_sky_order1_catalog.moc.fill = MagicMock()
    _, ax = small_sky_order1_catalog.plot_moc()
    small_sky_order1_catalog.moc.fill.assert_called_once()
    assert small_sky_order1_catalog.moc.fill.call_args[0][0] is ax
    wcs = ax.wcs
    assert small_sky_order1_catalog.moc.fill.call_args[0][1] is wcs



"""Container class to hold primary-to-join partition metadata"""

from __future__ import annotations

from pathlib import Path

import pandas as pd
from upath import UPath

from hats.catalog.partition_info import PartitionInfo
from hats.io import file_io, paths
from hats.pixel_math.healpix_pixel import HealpixPixel


class PartitionJoinInfo:
    """Association catalog metadata with which partitions matches occur in the join"""

    PRIMARY_ORDER_COLUMN_NAME = "Norder"
    PRIMARY_PIXEL_COLUMN_NAME = "Npix"
    JOIN_ORDER_COLUMN_NAME = "join_Norder"
    JOIN_PIXEL_COLUMN_NAME = "join_Npix"

    COLUMN_NAMES = [
        PRIMARY_ORDER_COLUMN_NAME,
        PRIMARY_PIXEL_COLUMN_NAME,
        JOIN_ORDER_COLUMN_NAME,
        JOIN_PIXEL_COLUMN_NAME,
    ]

    def __init__(self, join_info_df: pd.DataFrame, catalog_base_dir: str = None) -> None:
        self.data_frame = join_info_df
        self.catalog_base_dir = catalog_base_dir
        self._check_column_names()

    def _check_column_names(self):
        for column in self.COLUMN_NAMES:
            if column not in self.data_frame.columns:
                raise ValueError(f"join_info_df does not contain column {column}")

    def primary_to_join_map(self) -> dict[HealpixPixel, list[HealpixPixel]]:
        """Generate a map from a single primary pixel to one or more pixels in the join catalog.

        Lots of cute comprehension is happening here, so watch out!
        We create tuple of (primary order/pixel) and [array of tuples of (join order/pixel)]

        Returns:
            dictionary mapping (primary order/pixel) to [array of (join order/pixel)]
        """
        primary_map = self.data_frame.groupby(
            [self.PRIMARY_ORDER_COLUMN_NAME, self.PRIMARY_PIXEL_COLUMN_NAME], group_keys=True
        )
        primary_to_join = [
            (
                HealpixPixel(int(primary_pixel[0]), int(primary_pixel[1])),
                [
                    HealpixPixel(int(object_elem[0]), int(object_elem[1]))
                    for object_elem in join_group.dropna().to_numpy().T[2:4].T
                ],
            )
            for primary_pixel, join_group in primary_map
        ]
        ## Treat the array of tuples as a dictionary.
        primary_to_join = dict(primary_to_join)
        return primary_to_join

    def write_to_csv(self, catalog_path: str | Path | UPath | None = None):
        """Write all partition data to CSV files.

        Two files will be written:

            - partition_info.csv - covers all primary catalog pixels, and should match the file structure
            - partition_join_info.csv - covers all pairwise relationships between primary and
              join catalogs.

        Args:
            catalog_path: path to the directory where the
                `partition_join_info.csv` file will be written

        Raises:
            ValueError: if no path is provided, and could not be inferred.
        """
        if catalog_path is None:
            if self.catalog_base_dir is None:
                raise ValueError("catalog_path is required if info was not loaded from a directory")
            catalog_path = self.catalog_base_dir

        partition_join_info_file = paths.get_partition_join_info_pointer(catalog_path)
        file_io.write_dataframe_to_csv(self.data_frame, partition_join_info_file, index=False)

        primary_pixels = self.primary_to_join_map().keys()
        partition_info_pointer = paths.get_partition_info_pointer(catalog_path)
        partition_info = PartitionInfo.from_healpix(primary_pixels)
        partition_info.write_to_file(partition_info_file=partition_info_pointer)

    @classmethod
    def read_from_dir(cls, catalog_base_dir: str | Path | UPath | None = None) -> PartitionJoinInfo:
        """Read partition join info from a partition_join_info file within a hats directory.

        Args:
            catalog_base_dir: path to the root directory of the catalog

        Returns:
            A `PartitionJoinInfo` object with the data from the file

        Raises:
            FileNotFoundError: if the desired file is found in the catalog_base_dir
        """
        partition_join_info_file = paths.get_partition_join_info_pointer(catalog_base_dir)
        if file_io.does_file_or_directory_exist(partition_join_info_file):
            pixel_frame = PartitionJoinInfo._read_from_csv(partition_join_info_file)
        else:
            raise FileNotFoundError(
                f"partition join info file is required in catalog directory {catalog_base_dir}"
            )
        return cls(pixel_frame, catalog_base_dir)

    @classmethod
    def read_from_csv(cls, partition_join_info_file: str | Path | UPath) -> PartitionJoinInfo:
        """Read partition join info from a `partition_join_info.csv` file to create an object

        Args:
            partition_join_info_file (UPath): path to the `partition_join_info.csv` file

        Returns:
            A `PartitionJoinInfo` object with the data from the file
        """
        return cls(cls._read_from_csv(partition_join_info_file))

    @classmethod
    def _read_from_csv(cls, partition_join_info_file: str | Path | UPath) -> pd.DataFrame:
        """Read partition join info from a `partition_join_info.csv` file to create an object

        Args:
            partition_join_info_file (UPath): path to the `partition_join_info.csv` file

        Returns:
            A `PartitionJoinInfo` object with the data from the file
        """
        if not file_io.does_file_or_directory_exist(partition_join_info_file):
            raise FileNotFoundError(
                f"No partition join info found where expected: {str(partition_join_info_file)}"
            )

        return file_io.load_csv_to_pandas(partition_join_info_file)



"""Generate a molleview map with the pixel densities of the catalog

NB: Testing validity of generated plots is currently not tested in our unit test suite.
"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Type

import astropy.units as u
import astropy.wcs
import cdshealpix
import numpy as np
from astropy.coordinates import ICRS, Angle, SkyCoord
from astropy.units import Quantity
from astropy.visualization.wcsaxes import WCSAxes
from astropy.visualization.wcsaxes.frame import BaseFrame, EllipticalFrame
from astropy.wcs.utils import pixel_to_skycoord, skycoord_to_pixel
from matplotlib import pyplot as plt
from matplotlib.collections import PathCollection
from matplotlib.colors import Colormap, Normalize
from matplotlib.figure import Figure
from matplotlib.path import Path
from mocpy import MOC, WCS
from mocpy.moc.plot.culling_backfacing_cells import backface_culling
from mocpy.moc.plot.utils import _set_wcs

import hats.pixel_math.healpix_shim as hp
from hats.io import file_io, paths
from hats.pixel_math import HealpixPixel
from hats.pixel_tree.moc_filter import perform_filter_by_moc
from hats.pixel_tree.pixel_tree import PixelTree

if TYPE_CHECKING:
    from hats.catalog import Catalog
    from hats.catalog.healpix_dataset.healpix_dataset import HealpixDataset


def _read_point_map(catalog_base_dir):
    """Read the object spatial distribution information from a healpix FITS file.

    Args:
        catalog_base_dir: path to a catalog
    Returns:
        one-dimensional numpy array of long integers where the value at each index
        corresponds to the number of objects found at the healpix pixel.
    """
    map_file_pointer = paths.get_point_map_file_pointer(catalog_base_dir)
    return file_io.read_fits_image(map_file_pointer)


def plot_density(catalog: Catalog, *, plot_title: str | None = None, order=None, unit=None, **kwargs):
    """Create a visual map of the density of input points of a catalog on-disk.
    Args:
        catalog (`hats.catalog.Catalog`) Catalog to display
        plot_title (str): Optional title for the plot
        order (int): Optionally reduce the display healpix order, and aggregate smaller tiles.
        kwargs: Additional args to pass to `plot_healpix_map`
    """
    if catalog is None or not catalog.on_disk:
        raise ValueError("on disk catalog required for point-wise visualization")
    point_map = _read_point_map(catalog.catalog_base_dir)
    map_order = hp.npix2order(len(point_map))

    if order is not None:
        if order > map_order:
            raise ValueError(f"plotting order should be less than stored density map order ({map_order})")
        ## Create larger pixel sums from the constituent pixels.
        point_map = point_map.reshape(hp.order2npix(order), -1).sum(axis=1)
    else:
        order = map_order
    if unit is None:
        unit = u.deg * u.deg

    pix_area = hp.order2pixarea(order, unit=unit)

    point_map = point_map / pix_area
    default_title = f"Angular density of catalog {catalog.catalog_name}"
    fig, ax = plot_healpix_map(
        point_map, title=default_title if plot_title is None else plot_title, cbar=False, **kwargs
    )
    col = ax.collections[0]
    plt.colorbar(
        col,
        label=f"count / {unit} sq",
    )
    return fig, ax


def plot_pixels(catalog: HealpixDataset, plot_title: str | None = None, **kwargs):
    """Create a visual map of the pixel density of the catalog.

    Args:
        catalog (`hats.catalog.Catalog`) Catalog to display
        plot_title (str): Optional title for the plot
        kwargs: Additional args to pass to `plot_healpix_map`
    """
    pixels = catalog.get_healpix_pixels()
    default_title = f"Catalog pixel density map - {catalog.catalog_name}"
    return plot_pixel_list(
        pixels=pixels,
        plot_title=default_title if plot_title is None else plot_title,
        **kwargs,
    )


def plot_pixel_list(
    pixels: list[HealpixPixel], plot_title: str = "", projection="MOL", color_by_order=True, **kwargs
):
    """Create a visual map of the pixel density of a list of pixels.

    Args:
        pixels: healpix pixels (order and pixel number) to visualize
        plot_title (str): heading for the plot
        projection (str): The projection to use. Available projections listed at
            https://docs.astropy.org/en/stable/wcs/supported_projections.html
        color_by_order (bool): Whether to color the pixels by their order. True by default.
        kwargs: Additional args to pass to `plot_healpix_map`
    """
    orders = np.array([p.order for p in pixels])
    ipix = np.array([p.pixel for p in pixels])
    order_map = orders.copy()
    fig, ax = plot_healpix_map(
        order_map, projection=projection, title=plot_title, ipix=ipix, depth=orders, cbar=False, **kwargs
    )
    col = ax.collections[0]
    if color_by_order:
        col_array = col.get_array()
        plt.colorbar(
            col,
            boundaries=np.arange(np.min(col_array) - 0.5, np.max(col_array) + 0.6, 1),
            ticks=np.arange(np.min(col_array), np.max(col_array) + 1),
            label="HEALPix Order",
        )
    else:
        col.set(cmap=None, norm=None, array=None)
    return fig, ax


def plot_moc(
    moc: MOC,
    *,
    projection: str = "MOL",
    title: str = "",
    fov: Quantity | tuple[Quantity, Quantity] = None,
    center: SkyCoord | None = None,
    wcs: astropy.wcs.WCS = None,
    frame_class: Type[BaseFrame] | None = None,
    ax: WCSAxes | None = None,
    fig: Figure | None = None,
    **kwargs,
) -> tuple[Figure, WCSAxes]:
    """Plots a moc

    By default, a new matplotlib figure and axis will be created, and the projection will be a Molleweide
    projection across the whole sky.

    Args:
        moc (mocpy.MOC): MOC to plot
        projection (str): The projection to use in the WCS. Available projections listed at
            https://docs.astropy.org/en/stable/wcs/supported_projections.html
        title (str): The title of the plot
        fov (Quantity or Sequence[Quantity, Quantity] | None): The Field of View of the WCS. Must be an
            astropy Quantity with an angular unit, or a tuple of quantities for different longitude and \
            latitude FOVs (Default covers the full sky)
        center (SkyCoord | None): The center of the projection in the WCS (Default: SkyCoord(0, 0))
        wcs (WCS | None): The WCS to specify the projection of the plot. If used, all other WCS parameters
            are ignored and the parameters from the WCS object is used.
        frame_class (Type[BaseFrame] | None): The class of the frame for the WCSAxes to be initialized with.
            if the `ax` kwarg is used, this value is ignored (By Default uses EllipticalFrame for full
            sky projection. If FOV is set, RectangularFrame is used)
        ax (WCSAxes | None): The matplotlib axes to plot onto. If None, an axes will be created to be used. If
            specified, the axes must be an astropy WCSAxes, and the `wcs` parameter must be set with the WCS
            object used in the axes. (Default: None)
        fig (Figure | None): The matplotlib figure to add the axes to. If None, one will be created, unless
            ax is specified (Default: None)
        **kwargs: Additional kwargs to pass to `mocpy.MOC.fill`

    Returns:
        Tuple[Figure, WCSAxes] - The figure and axes used to plot the healpix map
    """
    fig, ax, wcs = initialize_wcs_axes(
        projection=projection,
        fov=fov,
        center=center,
        wcs=wcs,
        frame_class=frame_class,
        ax=ax,
        fig=fig,
        figsize=(9, 5),
    )

    mocpy_args = {"alpha": 0.5, "fill": True, "color": "teal"}
    mocpy_args.update(**kwargs)

    moc.fill(ax, wcs, **mocpy_args)

    ax.coords[0].set_format_unit("deg")

    plt.grid()
    plt.ylabel("Dec")
    plt.xlabel("RA")
    plt.title(title)
    return fig, ax


def get_fov_moc_from_wcs(wcs: WCS) -> MOC | None:
    """Returns a MOC that matches the plot window defined by a WCS

    Modified from mocpy.moc.plot.utils.build_plotting_moc

    Args:
        wcs (astropy.WCS): The wcs object with the plot's projection

    Returns:
        The moc which defines the area of the sky that would be visible in a WCSAxes with the given WCS
    """
    # Get the MOC delimiting the FOV polygon
    width_px = int(wcs.wcs.crpix[0] * 2.0)  # Supposing the wcs is centered in the axis
    height_px = int(wcs.wcs.crpix[1] * 2.0)

    # Compute the sky coordinate path delimiting the viewport.
    # It consists of a closed polygon of (4 - 1)*4 = 12 vertices
    x_px = np.linspace(0, width_px, 4)
    y_px = np.linspace(0, height_px, 4)

    x, y = np.meshgrid(x_px, y_px)

    x_px = np.append(x[0, :-1], x[:-1, -1])
    x_px = np.append(x_px, x[-1, 1:][::-1])
    x_px = np.append(x_px, x[:-1, 0])

    y_px = np.append(y[0, :-1], y[:-1, -1])
    y_px = np.append(y_px, y[-1, :-1])
    y_px = np.append(y_px, y[1:, 0][::-1])

    # Disable the output of warnings when encoutering NaNs.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Inverse projection from pixel coordinate space to the world coordinate space
        viewport = pixel_to_skycoord(x_px, y_px, wcs)
        # If one coordinate is a NaN we exit the function and do not go further
        ra_deg, dec_deg = viewport.icrs.ra.deg, viewport.icrs.dec.deg

    if np.isnan(ra_deg).any() or np.isnan(dec_deg).any():
        return None

    # Create a rough MOC (depth=3 is sufficient) from the viewport
    moc_viewport = MOC.from_polygon_skycoord(viewport, max_depth=3)
    return moc_viewport


def cull_to_fov(depth_ipix_d: dict[int, tuple[np.ndarray, np.ndarray]], wcs):
    """Culls a mapping of ipix to values to pixels that are inside the plot window defined by a WCS

    Modified from mocpy.moc.plot.utils.build_plotting_moc

    Any pixels too small are merged to a lower order, with the map values within a lower order pixel being
    sampled

    Args:
        depth_ipix_d (Dict[int, Tuple[np.ndarray, np.ndarray]]): Map of HEALPix order to a tuple of 2 arrays
            (the ipix array of pixel numbers in NESTED ordering, and the values of the pixels)
        wcs (astropy.WCS): The wcs object with the plot's projection

    Returns:
        A new map with the same datatype of depth_ipix_d, with any pixels outside the plot's FOV removed,
        and any pixels too small merged with their map values subsampled.
    """

    depth_ipix_d = _merge_too_small_pixels(depth_ipix_d, wcs)

    moc_viewport = get_fov_moc_from_wcs(wcs)

    if moc_viewport is None:
        return depth_ipix_d

    output_dict = {}

    # The moc to plot is the INPUT_MOC & MOC_VIEWPORT. For small FOVs this can reduce
    # a lot the time to draw the MOC along with its borders.
    for d, (ip, vals) in depth_ipix_d.items():
        ip_argsort = np.argsort(ip)
        ip_sorted = ip[ip_argsort]
        pixel_tree = PixelTree(np.vstack([ip_sorted, ip_sorted + 1]).T, order=d)
        ip_viewport_mask = perform_filter_by_moc(
            pixel_tree.to_depth29_ranges(), moc_viewport.to_depth29_ranges
        )
        output_dict[d] = (ip_sorted[ip_viewport_mask], vals[ip_argsort][ip_viewport_mask])
    return output_dict


def _merge_too_small_pixels(depth_ipix_d: dict[int, tuple[np.ndarray, np.ndarray]], wcs):
    """Merges any pixels too small in a map to a lower order, with the map values within a lower order pixel
    being sampled"""
    # Get the WCS cdelt giving the deg.px^(-1) resolution.
    cdelt = wcs.wcs.cdelt
    # Convert in rad.px^(-1)
    cdelt = np.abs((2 * np.pi / 360) * cdelt[0])
    # Get the minimum depth such as the resolution of a cell is contained in 1px.
    depth_res = int(np.floor(np.log2(np.sqrt(np.pi / 3) / cdelt)))
    depth_res = max(depth_res, 0)

    max_depth = max(depth_ipix_d.keys())

    # Combine healpix pixels smaller than 1px in the plot
    if max_depth > depth_res:
        warnings.warn(
            "This plot contains HEALPix pixels smaller than a pixel of the plot. Some values may be lost"
        )
        new_ipix_d = {}
        for d, (ip, vals) in depth_ipix_d.items():
            if d <= depth_res:
                new_ipix_d[d] = (ip, vals)
            else:
                ipix_depth_res = ip >> (2 * (d - depth_res))
                # Get the unique pixels at the maximum depth resolution
                unique_ipix, unique_indices = np.unique(ipix_depth_res, return_index=True)
                # Get the first values from the map for each lower order pixel
                vals_depth_res = vals[unique_indices]
                if depth_res not in new_ipix_d:
                    new_ipix_d[depth_res] = (unique_ipix, vals_depth_res)
                else:
                    # combine with existing pixels if they exist
                    ipix_depth_res = np.concatenate([new_ipix_d[depth_res][0], unique_ipix])
                    vals_depth_res = np.concatenate([new_ipix_d[depth_res][1], vals_depth_res])
                    ip_argsort = np.argsort(ipix_depth_res)
                    new_ipix_d[depth_res] = (ipix_depth_res[ip_argsort], vals_depth_res[ip_argsort])
        depth_ipix_d = new_ipix_d
    return depth_ipix_d


def cull_from_pixel_map(depth_ipix_d: dict[int, tuple[np.ndarray, np.ndarray]], wcs, max_split_depth=7):
    """Modified from mocpy.moc.plot.culling_backfacing_cells.from_moc

    Create a new MOC that do not contain the HEALPix cells that are backfacing the projection.

    Args:
        depth_ipix_d (Dict[int, Tuple[np.ndarray, np.ndarray]]): Map of HEALPix order to a tuple of 2 arrays
            (the ipix array of pixel numbers in NESTED ordering, and the values of the pixels)
        wcs (astropy.WCS): The wcs object with the plot's projection
        max_split_depth: the max depth to split backfacing cells to

    Returns:
        A new map with the same datatype of depth_ipix_d, with backfacing cells split into higher order
    """
    depths = list(depth_ipix_d.keys())
    min_depth = min(depths)
    max_depth = max(depths)
    ipixels, vals = depth_ipix_d[min_depth]

    # Split the cells located at the border of the projection
    # until at least the max_split_depth
    max_split_depth = max(max_split_depth, max_depth)

    ipix_d = {}
    for depth in range(min_depth, max_split_depth + 1):
        # for each depth, check if pixels are too large, or wrap around projection, and split into pixels at
        # higher order
        ipix_lon, ipix_lat = cdshealpix.vertices(ipixels, depth)

        ipix_lon = ipix_lon[:, [2, 3, 0, 1]]
        ipix_lat = ipix_lat[:, [2, 3, 0, 1]]
        ipix_vertices = SkyCoord(ipix_lon, ipix_lat, frame=ICRS())

        # Projection on the given WCS
        xp, yp = skycoord_to_pixel(coords=ipix_vertices, wcs=wcs)
        _, _, frontface_id = backface_culling(xp, yp)

        # Get the pixels which are backfacing the projection
        backfacing_ipix = ipixels[~frontface_id]  # backfacing
        backfacing_vals = vals[~frontface_id]
        frontface_ipix = ipixels[frontface_id]
        frontface_vals = vals[frontface_id]

        ipix_d.update({depth: (frontface_ipix, frontface_vals)})

        too_large_ipix = backfacing_ipix
        too_large_vals = backfacing_vals

        next_depth = depth + 1

        # get next depth if there is one, or use empty array as default
        ipixels = np.array([], dtype=ipixels.dtype)
        vals = np.array([], dtype=vals.dtype)

        if next_depth in depth_ipix_d:
            ipixels, vals = depth_ipix_d[next_depth]

        # split too large ipix into next order, with each child getting the same map value as parent

        too_large_child_ipix = np.repeat(too_large_ipix << 2, 4) + np.tile(
            np.array([0, 1, 2, 3]), len(too_large_ipix)
        )
        too_large_child_vals = np.repeat(too_large_vals, 4)

        ipixels = np.concatenate((ipixels, too_large_child_ipix))
        vals = np.concatenate((vals, too_large_child_vals))

    return ipix_d


def compute_healpix_vertices(depth, ipix, wcs, step=1):
    """Compute HEALPix vertices.

    Modified from mocpy.moc.plot.fill.compute_healpix_vertices

    Parameters
    ----------
    depth : int
        The depth of the HEALPix cells.
    ipix : `numpy.ndarray`
        The HEALPix cell index given as a `np.uint64` numpy array.
    wcs : `astropy.wcs.WCS`
        A WCS projection

    Returns
    -------
    path_vertices, codes : (`np.array`, `np.array`)
    """

    depth = int(depth)

    ipix_lon, ipix_lat = cdshealpix.vertices(ipix, depth, step=step)
    indices = np.concatenate([np.arange(2 * step, 4 * step), np.arange(2 * step)])

    ipix_lon = ipix_lon[:, indices]
    ipix_lat = ipix_lat[:, indices]
    ipix_boundaries = SkyCoord(ipix_lon, ipix_lat, frame=ICRS())
    # Projection on the given WCS
    xp, yp = skycoord_to_pixel(ipix_boundaries, wcs=wcs)

    raw_cells = [np.vstack((xp[:, i], yp[:, i])).T for i in range(4 * step)]

    cells = np.hstack(raw_cells + [np.zeros((raw_cells[0].shape[0], 2))])

    path_vertices = cells.reshape(((4 * step + 1) * raw_cells[0].shape[0], 2))
    single_code = np.array([Path.MOVETO] + [Path.LINETO] * (step * 4 - 1) + [Path.CLOSEPOLY])

    codes = np.tile(single_code, raw_cells[0].shape[0])

    return path_vertices, codes


def plot_healpix_map(
    healpix_map: np.ndarray,
    *,
    projection: str = "MOL",
    title: str = "",
    cmap: str | Colormap = "viridis",
    norm: Normalize | None = None,
    ipix: np.ndarray | None = None,
    depth: np.ndarray | None = None,
    cbar: bool = True,
    fov: Quantity | tuple[Quantity, Quantity] = None,
    center: SkyCoord | None = None,
    wcs: astropy.wcs.WCS = None,
    frame_class: Type[BaseFrame] | None = None,
    ax: WCSAxes | None = None,
    fig: Figure | None = None,
    **kwargs,
):
    """Plot a map of HEALPix pixels to values as a colormap across a projection of the sky

    Plots the given healpix pixels on a spherical projection defined by a WCS. Colors each pixel based on the
    corresponding value in a map.

    The map can be across all healpix pixels at a given order, or specify a subset of healpix pixels with the
    `ipix` and `depth` parameters.

    By default, a new matplotlib figure and axis will be created, and the projection will be a Molleweide
    projection across the whole sky.

    Additional kwargs will be passed to the creation of a matplotlib `PathCollection` object, which is the
    artist that draws the tiles.

    Args:
        healpix_map (np.ndarray): Array of map values for the healpix tiles. If ipix and depth are not
            specified, the length of this array will be used to determine the healpix order, and will plot
            healpix pixels with pixel index corresponding to the array index in NESTED ordering. If ipix and
            depth are specified, all arrays must be of the same length, and the pixels specified by the
            ipix and depth arrays will be plotted with their values specified in the healpix_map array.
        projection (str): The projection to use in the WCS. Available projections listed at
            https://docs.astropy.org/en/stable/wcs/supported_projections.html
        title (str): The title of the plot
        cmap (str | Colormap): The matplotlib colormap to plot with
        norm (Normalize | None): The matplotlib normalization to plot with
        ipix (np.ndarray | None): Array of HEALPix NESTED pixel indices. Must be used with depth, and arrays
            must be the same length
        depth (np.ndarray | None): Array of HEALPix pixel orders. Must be used with ipix, and arrays
            must be the same length
        cbar (bool): If True, includes a color bar in the plot (Default: True)
        fov (Quantity or Sequence[Quantity, Quantity] | None): The Field of View of the WCS. Must be an
            astropy Quantity with an angular unit, or a tuple of quantities for different longitude and \
            latitude FOVs (Default covers the full sky)
        center (SkyCoord | None): The center of the projection in the WCS (Default: SkyCoord(0, 0))
        wcs (WCS | None): The WCS to specify the projection of the plot. If used, all other WCS parameters
            are ignored and the parameters from the WCS object is used.
        frame_class (Type[BaseFrame] | None): The class of the frame for the WCSAxes to be initialized with.
            if the `ax` kwarg is used, this value is ignored (By Default uses EllipticalFrame for full
            sky projection. If FOV is set, RectangularFrame is used)
        ax (WCSAxes | None): The matplotlib axes to plot onto. If None, an axes will be created to be used. If
            specified, the axes must be an astropy WCSAxes, and the `wcs` parameter must be set with the WCS
            object used in the axes. (Default: None)
        fig (Figure | None): The matplotlib figure to add the axes to. If None, one will be created, unless
            ax is specified (Default: None)
        **kwargs: Additional kwargs to pass to creating the matplotlib `PathCollection` artist

    Returns:
        Tuple[Figure, WCSAxes] - The figure and axes used to plot the healpix map
    """
    if ipix is None or depth is None:
        order = int(np.ceil(np.log2(len(healpix_map) / 12) / 2))
        ipix = np.arange(len(healpix_map))
        depth = np.full(len(healpix_map), fill_value=order)

    fig, ax, wcs = initialize_wcs_axes(
        projection=projection,
        fov=fov,
        center=center,
        wcs=wcs,
        frame_class=frame_class,
        ax=ax,
        fig=fig,
        figsize=(10, 5),
    )

    _plot_healpix_value_map(ipix, depth, healpix_map, ax, wcs, cmap=cmap, norm=norm, cbar=cbar, **kwargs)
    plt.grid()
    plt.ylabel("Dec")
    plt.xlabel("RA")
    plt.title(title)
    return fig, ax


def initialize_wcs_axes(
    projection: str = "MOL",
    fov: Quantity | tuple[Quantity, Quantity] = None,
    center: SkyCoord | None = None,
    wcs: astropy.wcs.WCS = None,
    frame_class: Type[BaseFrame] | None = None,
    ax: WCSAxes | None = None,
    fig: Figure | None = None,
    **kwargs,
):
    """Initializes matplotlib Figure and WCSAxes if they do not exist

    Args:
        projection (str): The projection to use in the WCS. Available projections listed at
            https://docs.astropy.org/en/stable/wcs/supported_projections.html
        fov (Quantity or Sequence[Quantity, Quantity] | None): The Field of View of the WCS. Must be an
            astropy Quantity with an angular unit, or a tuple of quantities for different longitude and \
            latitude FOVs (Default covers the full sky)
        center (SkyCoord | None): The center of the projection in the WCS (Default: SkyCoord(0, 0))
        wcs (WCS | None): The WCS to specify the projection of the plot. If used, all other WCS parameters
            are ignored and the parameters from the WCS object is used.
        frame_class (Type[BaseFrame] | None): The class of the frame for the WCSAxes to be initialized with.
            if the `ax` kwarg is used, this value is ignored (By Default uses EllipticalFrame for full
            sky projection. If FOV is set, RectangularFrame is used)
        ax (WCSAxes | None): The matplotlib axes to plot onto. If None, the current axes will be used.
            If the current axes is not the correct WCSAxes type, a new figure and  axes will be created to be
            used. If specified, the axes must be an astropy WCSAxes, and the `wcs` parameter will be ignored
            and the wcs of the axes used. (Default: None)
        fig (Figure | None): The matplotlib figure to add the axes to. If None, the current figure will be
            used, unless ax is specified, in which case the figure of the ax will be used. If there is no
            current figure, one will be created. (Default: None)
        kwargs: additional kwargs to pass to figure initialization
    """
    if fig is None:
        if ax is not None:
            # Use figure from axes if ax provided
            fig = ax.get_figure()
        else:
            if len(plt.get_fignums()) == 0:
                # create new figure if no existing figure
                fig = plt.figure(**kwargs)
            else:
                # use current figure if exists
                fig = plt.gcf()
    if frame_class is None and fov is None and wcs is None:
        frame_class = EllipticalFrame
    if fov is None:
        fov = (320 * u.deg, 160 * u.deg)
    if center is None:
        center = SkyCoord(0, 0, unit="deg", frame="icrs")
    if ax is None:
        if len(fig.axes) > 0:
            ax = plt.gca()
            if isinstance(ax, WCSAxes):
                # Use current axes if axes exists in figure and current axes is correct type
                wcs = ax.wcs
                return fig, ax, wcs
            # Plot onto new axes on new figure if current axes is not correct type
            warnings.warn("Current axes is not of correct type WCSAxes. A new figure and axes will be used.")
            fig = plt.figure(**kwargs)
        if wcs is None:
            # Initialize wcs with params if no WCS provided
            wcs = WCS(
                fig,
                fov=fov,
                center=center,
                coordsys="icrs",
                rotation=Angle(0, u.deg),
                projection=projection,
            ).w
        # If no axes provided and no valid current axes then create new axes with the right projection
        ax = fig.add_subplot(1, 1, 1, projection=wcs, frame_class=frame_class)
    elif not isinstance(ax, WCSAxes):
        # Error if provided axes is of wrong type
        raise ValueError("ax is not of correct type WCSAxes")
    else:
        # Use WCS from provided axes if provided axes is correct type
        wcs = ax.wcs
    return fig, ax, wcs


def _plot_healpix_value_map(ipix, depth, values, ax, wcs, cmap="viridis", norm=None, cbar=True, **kwargs):
    """Perform the plotting of a healpix pixel map."""

    # create dict mapping depth to ipix and values
    depth_ipix_d = {}

    for d in np.unique(depth):
        mask = depth == d
        depth_ipix_d[d] = (ipix[mask], values[mask])

    # cull backfacing and out of fov cells
    fov_culled_d = cull_to_fov(depth_ipix_d, wcs)
    culled_d = cull_from_pixel_map(fov_culled_d, wcs)

    # Generate Paths for each pixel and add to ax
    plt_paths = []
    cum_vals = []
    for d, (ip, vals) in culled_d.items():
        step = 1 if d >= 3 else 2 ** (3 - d)
        vertices, codes = compute_healpix_vertices(depth=d, ipix=ip, wcs=wcs, step=step)
        for i in range(len(ip)):
            plt_paths.append(
                Path(
                    vertices[(4 * step + 1) * i : (4 * step + 1) * (i + 1)],
                    codes[(4 * step + 1) * i : (4 * step + 1) * (i + 1)],
                )
            )
        cum_vals.append(vals)
    col = PathCollection(plt_paths, cmap=cmap, norm=norm, **kwargs)
    col.set_array(np.concatenate(cum_vals))
    ax.add_collection(col)

    # Add color bar
    if cbar:
        plt.colorbar(col)

    # Set projection
    _set_wcs(ax, wcs)

    ax.coords[0].set_format_unit("deg")



from __future__ import annotations

from enum import Enum

import numpy as np

import hats.pixel_math.healpix_shim as hp


class ValidatorsErrors(str, Enum):
    """Error messages for the coordinate validators"""

    INVALID_DEC = "declination must be in the -90.0 to 90.0 degree range"
    INVALID_RADIUS = "cone radius must be positive"
    INVALID_NUM_VERTICES = "polygon must contain a minimum of 3 vertices"
    DUPLICATE_VERTICES = "polygon has duplicated vertices"
    DEGENERATE_POLYGON = "polygon is degenerate"
    INVALID_RADEC_RANGE = "invalid ra or dec range"
    INVALID_COORDS_SHAPE = "invalid coordinates shape"
    INVALID_CONCAVE_SHAPE = "polygon must be convex"


def validate_radius(radius_arcsec: float):
    """Validates that a cone search radius is positive

    Args:
        radius_arcsec (float): The cone radius, in arcseconds

    Raises:
        ValueError: if radius is non-positive
    """
    if radius_arcsec <= 0:
        raise ValueError(ValidatorsErrors.INVALID_RADIUS.value)


def validate_declination_values(dec: float | list[float]):
    """Validates that declination values are in the [-90,90] degree range

    Args:
        dec (float | List[float]): The declination values to be validated

    Raises:
        ValueError: if declination values are not in the [-90,90] degree range
    """
    dec_values = np.array(dec)
    lower_bound, upper_bound = -90.0, 90.0
    if not np.all((dec_values >= lower_bound) & (dec_values <= upper_bound)):
        raise ValueError(ValidatorsErrors.INVALID_DEC.value)


def validate_polygon(vertices: list[tuple[float, float]]):
    """Checks if the polygon contain a minimum of three vertices, that they are
    unique and that the polygon does not fall on a great circle.

    Args:
        vertices (list[tuple[float,float]]): The list of vertice coordinates for
            the polygon, (ra, dec), in degrees.

    Raises:
        ValueError: exception if the polygon is invalid.
    """
    vertices = np.array(vertices)
    if vertices.shape[1] != 2:
        raise ValueError(ValidatorsErrors.INVALID_COORDS_SHAPE.value)
    _, dec = vertices.T
    validate_declination_values(dec)
    if len(vertices) < 3:
        raise ValueError(ValidatorsErrors.INVALID_NUM_VERTICES.value)
    if len(vertices) != len(np.unique(vertices, axis=0)):
        raise ValueError(ValidatorsErrors.DUPLICATE_VERTICES.value)
    check_polygon_is_valid(vertices)


def check_polygon_is_valid(vertices: np.ndarray):
    """Check if the polygon has no degenerate corners and it is convex.

    Args:
        vertices (np.ndarray): The polygon vertices, in cartesian coordinates
    """
    vertices_xyz = hp.ang2vec(*vertices.T)

    # Compute the normal between each pair of neighboring vertices
    second_vertices = np.roll(vertices_xyz, -1, axis=0)
    normals = np.cross(vertices_xyz, second_vertices)

    # Compute the dot products between each normal and a third neighboring vertex.
    # 'ij,ij->i' means we will multiply each normal vector with the corresponding
    # vector of the third vertex ('ij,ij': hence element-wise) and sum over each
    # column for each row '->i'.
    third_vertices = np.roll(second_vertices, -1, axis=0)
    dot_products = np.einsum("ij,ij->i", normals, third_vertices)

    if np.any(np.isclose(dot_products, 0)):
        raise ValueError(ValidatorsErrors.DEGENERATE_POLYGON.value)
    if not (np.all(dot_products > 0) or np.all(dot_products < 0)):
        raise ValueError(ValidatorsErrors.INVALID_CONCAVE_SHAPE.value)


def validate_box(ra: tuple[float, float], dec: tuple[float, float]):
    """Checks if ra and dec values are valid for the box search.

    - Both ranges for ra or dec must have been provided.
    - Ranges must be defined by a pair of values, in degrees.
    - Declination values must be unique, provided in ascending order, and
    belong to the [-90,90] degree range.

    Args:
        ra (Tuple[float, float]): Right ascension range, in degrees
        dec (Tuple[float, float]): Declination range, in degrees
    """
    invalid_range = False
    if ra is None or len(ra) != 2:
        invalid_range = True
    elif dec is None or len(dec) != 2 or dec[0] >= dec[1]:
        invalid_range = True
    if invalid_range:
        raise ValueError(ValidatorsErrors.INVALID_RADEC_RANGE.value)
    validate_declination_values(dec)



import astropy.units as u
import cdshealpix
import numpy as np
import pytest
from astropy.coordinates import Latitude, Longitude
from numpy.testing import assert_allclose, assert_array_equal

from hats.pixel_math import healpix_shim as hps


def test_avgsize2mindist2avgsize():
    """Test that avgsize2mindist is inverse of mindist2avgsize"""
    avgsize = np.logspace(-5, 5, 21)
    assert_allclose(hps.mindist2avgsize(hps.avgsize2mindist(avgsize)), avgsize, rtol=1e-12)


def test_mindist2avgsize2mindist():
    """Test that mindist2avgsize is inverse of avgsize2mindist"""
    mindist = np.logspace(-3.2, 2.8, 21)
    assert_allclose(hps.avgsize2mindist(hps.mindist2avgsize(mindist)), mindist, rtol=1e-12)


def test_order2avgsize2order():
    """Test that avgsize2order is inverse of hps.nside2resol(hps.order2nside, arcmin=True)"""
    order = np.arange(20)
    assert_array_equal(hps.avgsize2order(hps.order2resol(order, arcmin=True)), order)


def test_margin2order():
    """Test margin2order for some pre-computed values"""
    margin_thr_arcmin = np.array([1 / 60, 10 / 60, 1, 5, 60])
    orders = np.array([17, 13, 11, 8, 5])
    assert_array_equal(hps.margin2order(margin_thr_arcmin), orders)


def test_order2mindist():
    """Test order2mindist for some pre-computed values"""
    orders = np.array([17, 13, 11, 8, 5])
    min_distances = np.array([0.01677, 0.268, 1.07, 8.588, 68.7])
    assert_allclose(hps.order2mindist(orders), min_distances, rtol=1e-2)

    assert_allclose(hps.order2mindist(17), 0.01677, rtol=1e-2)


def test_ang2vec():
    """Tests conversion of ra/dec to unit cartesian vectors"""
    ra = [230.14467816, 110.40507118, 9.41764689, 245.5553117]
    dec = [38.78080888, 17.09584081, -28.6352765, 5.41803306]
    ra_rad = np.radians(ra)
    dec_rad = np.radians(dec)
    x = np.cos(ra_rad) * np.cos(dec_rad)
    y = np.sin(ra_rad) * np.cos(dec_rad)
    z = np.sin(dec_rad)
    actual = np.asarray([x, y, z]).T
    assert_array_equal(actual, hps.ang2vec(ra, dec))


def test_npix2order():
    orders = [0, 1, 5, 10, 20, 29]
    npix = [12 * (4**order) for order in orders]
    test_orders = [hps.npix2order(x) for x in npix]
    assert test_orders == orders


def test_npix2order_invalid():
    npixs = [-10, 0, 11, 13, 47, 49, 100000, 100000000000000000]
    for npix in npixs:
        with pytest.raises(ValueError, match="Invalid"):
            hps.npix2order(npix)


def test_order2nside():
    orders = [0, 1, 5, 10, 20, 29]
    expected_nsides = [2**x for x in orders]
    test_nsides = [hps.order2nside(o) for o in orders]
    assert test_nsides == expected_nsides


def test_order2nside_invalid():
    orders = [-1000, -1, 30, 4000]
    for order in orders:
        with pytest.raises(ValueError, match="Invalid"):
            hps.order2nside(order)


def test_order2npix():
    orders = [0, 1, 5, 10, 20, 29]
    npix = [12 * (4**order) for order in orders]
    test_npix = [hps.order2npix(o) for o in orders]
    assert test_npix == npix


def test_order2npix_invalid():
    orders = [-1000, -1, 30, 4000]
    for order in orders:
        with pytest.raises(ValueError, match="Invalid"):
            hps.order2npix(order)


def test_order2pixarea():
    orders = [0, 1, 5, 10, 20, 29]
    npix = [12 * (4**order) for order in orders]
    pix_area_expected = [4 * np.pi / x for x in npix]
    pix_area_test = [hps.order2pixarea(order) for order in orders]
    assert pix_area_test == pix_area_expected


def test_order2pixarea_units():
    orders = [0, 1, 5, 10, 20, 29]
    npix = [12 * (4**order) for order in orders]
    pix_area_expected = [np.rad2deg(np.rad2deg(4 * np.pi / x)) for x in npix]
    pix_area_test = [hps.order2pixarea(order, degrees=True) for order in orders]
    assert pix_area_test == pix_area_expected

    pix_area_expected = [np.rad2deg(np.rad2deg(4 * np.pi / x)) * 3600 for x in npix]
    pix_area_test = [hps.order2pixarea(order, unit="arcmin2") for order in orders]
    assert_allclose(pix_area_test, pix_area_expected)

    pix_area_test = [hps.order2pixarea(order, unit=u.arcmin * u.arcmin) for order in orders]
    assert_allclose(pix_area_test, pix_area_expected)

    pix_area_expected = [np.rad2deg(np.rad2deg(4 * np.pi / x)) * 12960000 for x in npix]
    pix_area_test = [hps.order2pixarea(order, unit="arcsec2") for order in orders]
    assert_allclose(pix_area_test, pix_area_expected)

    with pytest.raises(ValueError, match="not convertible"):
        hps.order2pixarea(10, unit="arcmin")


def test_order2resol():
    orders = [0, 1, 5, 10, 20, 29]
    resol_expected = [np.sqrt(hps.order2pixarea(order)) for order in orders]
    resol_test = [hps.order2resol(order) for order in orders]
    assert resol_test == resol_expected


def test_order2resol_arcmin():
    orders = [0, 1, 5, 10, 20, 29]
    resol_expected = [np.rad2deg(np.sqrt(hps.order2pixarea(order))) * 60 for order in orders]
    resol_test = [hps.order2resol(order, arcmin=True) for order in orders]
    assert_allclose(resol_test, resol_expected)


def test_order2resol_degree():
    orders = [0, 1, 5, 10, 20, 29]
    resol_expected = [np.rad2deg(np.sqrt(hps.order2pixarea(order))) for order in orders]
    resol_test = [hps.order2resol(order, unit=u.deg) for order in orders]
    assert_allclose(resol_test, resol_expected)

    resol_test = [hps.order2resol(order, unit=u.degree) for order in orders]
    assert_allclose(resol_test, resol_expected)

    resol_test = [hps.order2resol(order, unit="deg") for order in orders]
    assert_allclose(resol_test, resol_expected)


def test_radec2pix_lonlat():
    orders = [0, 1, 5, 10, 20, 29]
    ras = np.arange(-180.0, 180.0, 10.0)
    decs = np.arange(-90.0, 90.0, 180 // len(ras))
    for order in orders:
        expected_pixels = cdshealpix.lonlat_to_healpix(
            Longitude(ras, unit="deg"), Latitude(decs, unit="deg"), order
        )
        pixels = hps.radec2pix(order, ras, decs)
        assert np.all(pixels == expected_pixels)


def test_radec2pix_lonlat_float32():
    orders = [0, 1, 5, 10, 20, 29]
    ras = np.arange(-180.0, 180.0, 10.0)
    ras_f = ras.astype(np.float32)
    decs = np.arange(-90.0, 90.0, 180 // len(ras))
    decs_f = decs.astype(np.float32)
    for order in orders:
        expected_pixels = cdshealpix.lonlat_to_healpix(
            Longitude(ras, unit="deg"), Latitude(decs, unit="deg"), order
        )
        # Verify that healpixshim can work with float32 versions
        # Fixes https://github.com/astronomy-commons/hats-import/issues/458
        pixels = hps.radec2pix(order, ras_f, decs_f)
        assert np.all(pixels == expected_pixels)


def test_radec2pix_invalid():
    orders = [0, 1, 5, 10, 20, 29]
    invalid_orders = [-1000, -1, 30, 40]
    ras = np.arange(-4000.0, 1000.0, 100.0)
    decs = np.arange(-1000.0, 1000.0, 2000.0 // len(ras))
    for order in invalid_orders:
        with pytest.raises(ValueError, match="Invalid"):
            hps.radec2pix(order, ras, decs)
    for order in orders:
        with pytest.raises(ValueError, match="angle"):
            hps.radec2pix(order, ras, decs)



from __future__ import annotations

import pandas as pd
import pyarrow as pa
from mocpy import MOC

from hats.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hats.catalog.dataset.table_properties import TableProperties
from hats.catalog.healpix_dataset.healpix_dataset import HealpixDataset
from hats.catalog.partition_info import PartitionInfo
from hats.pixel_math import HealpixPixel
from hats.pixel_tree.pixel_tree import PixelTree


class AssociationCatalog(HealpixDataset):
    """A HATS Catalog for enabling fast joins between two HATS catalogs

    Catalogs of this type are partitioned based on the partitioning of the left catalog.
    The `partition_join_info` metadata file specifies all pairs of pixels in the Association
    Catalog, corresponding to each pair of partitions in each catalog that contain rows to join.
    """

    def __init__(
        self,
        catalog_info: TableProperties,
        pixels: PartitionInfo | PixelTree | list[HealpixPixel],
        join_pixels: list | pd.DataFrame | PartitionJoinInfo,
        catalog_path=None,
        moc: MOC | None = None,
        schema: pa.Schema | None = None,
    ) -> None:
        super().__init__(catalog_info, pixels, catalog_path, moc=moc, schema=schema)
        self.join_info = self._get_partition_join_info_from_pixels(join_pixels)

    def get_join_pixels(self) -> pd.DataFrame:
        """Get join pixels listing all pairs of pixels from left and right catalogs that contain
        matching association rows

        Returns:
            pd.DataFrame with each row being a pair of pixels from the primary and join catalogs
        """
        return self.join_info.data_frame

    @staticmethod
    def _get_partition_join_info_from_pixels(
        join_pixels: list | pd.DataFrame | PartitionJoinInfo,
    ) -> PartitionJoinInfo:
        if isinstance(join_pixels, PartitionJoinInfo):
            return join_pixels
        if isinstance(join_pixels, pd.DataFrame):
            return PartitionJoinInfo(join_pixels)
        raise TypeError("join_pixels must be of type PartitionJoinInfo or DataFrame")



import pytest

from hats.catalog.dataset.collection_properties import CollectionProperties
from hats.io.file_io import write_string_to_file
from hats.io.file_io.file_io import load_text_file


def test_read_collection_from_file(small_sky_collection_dir):
    properties_from_file = CollectionProperties.read_from_dir(small_sky_collection_dir)

    expected_properties = CollectionProperties(
        name="small_sky_01",
        hats_primary_table_url="small_sky_order1",
        all_margins="small_sky_order1_margin",
        default_margin="small_sky_order1_margin",
        all_indexes="id small_sky_order1_id_index",
        default_index="id",
        obs_regime="Optical",
    )

    assert properties_from_file == expected_properties


def test_collection_properties_string():
    expected_properties = CollectionProperties(
        name="small_sky_01",
        hats_primary_table_url="small_sky_order1",
        obs_regime="Optical",
    )

    ## str representation should not include additional properties.
    assert (
        str(expected_properties)
        == """  name small_sky_01
  hats_primary_table_url small_sky_order1
"""
    )


def test_read_collection_list_parse(tmp_path):
    test_file_path = tmp_path / "collection.properties"

    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  all_margins=small_sky_order1_margin small_sky_order1_margin_10arcs
  all_indexes=
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    parsed_properties = CollectionProperties.read_from_dir(tmp_path)
    assert parsed_properties.all_margins == ["small_sky_order1_margin", "small_sky_order1_margin_10arcs"]
    assert parsed_properties.all_indexes is None

    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  all_margins=
  all_indexes=id small_sky_order1_id_index
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    parsed_properties = CollectionProperties.read_from_dir(tmp_path)
    assert parsed_properties.all_margins is None
    assert parsed_properties.all_indexes == {"id": "small_sky_order1_id_index"}

    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  all_margins=
  all_indexes=
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    parsed_properties = CollectionProperties.read_from_dir(tmp_path)

    parsed_properties.to_properties_file(tmp_path)

    contents = load_text_file(test_file_path)
    assert contents == [
        "#HATS Collection\n",
        "obs_collection=foo\n",
        "hats_primary_table_url=small_sky_order1\n",
    ]


def test_read_collection_read_errors(tmp_path):
    test_file_path = tmp_path / "collection.properties"

    with pytest.raises(FileNotFoundError):
        CollectionProperties.read_from_dir(tmp_path)

    ## Very basic content - should work.
    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    _ = CollectionProperties.read_from_dir(tmp_path)

    ## Unsupported extra property.
    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  hats_col_ra=ra
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    with pytest.raises(ValueError, match="hats_col_ra"):
        CollectionProperties.read_from_dir(tmp_path)

    ## Unmatched index map.
    file_content = """  name= foo
  hats_primary_table_url=small_sky_order1
  all_indexes=ra
"""
    write_string_to_file(test_file_path, file_content, encoding="utf-8")

    with pytest.raises(ValueError, match="all_indexes"):
        CollectionProperties.read_from_dir(tmp_path)


def test_read_collection_from_file_round_trip(small_sky_collection_dir, tmp_path):
    table_properties = CollectionProperties.read_from_dir(small_sky_collection_dir)
    table_properties.to_properties_file(tmp_path)
    round_trip_properties = CollectionProperties.read_from_dir(tmp_path)

    assert table_properties == round_trip_properties

    kwarg_properties = CollectionProperties(
        **round_trip_properties.model_dump(by_alias=False, exclude_none=True)
    )
    assert table_properties == kwarg_properties



"""Utility functions for handling parquet metadata files"""

from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow.dataset as pds
from upath import UPath

from hats.io import file_io, paths
from hats.io.file_io.file_pointer import get_upath
from hats.pixel_math.healpix_pixel import HealpixPixel
from hats.pixel_math.healpix_pixel_function import get_pixel_argsort


def write_parquet_metadata(
    catalog_path: str | Path | UPath,
    order_by_healpix=True,
    output_path: str | Path | UPath | None = None,
):
    """Generate parquet metadata, using the already-partitioned parquet files
    for this catalog.

    For more information on the general parquet metadata files, and why we write them, see
    https://arrow.apache.org/docs/python/parquet.html#writing-metadata-and-common-metadata-files

    Args:
        catalog_path (str): base path for the catalog
        order_by_healpix (bool): use False if the dataset is not to be reordered by
            breadth-first healpix pixel (e.g. secondary indexes)
        output_path (str): base path for writing out metadata files
            defaults to `catalog_path` if unspecified

    Returns:
        sum of the number of rows in the dataset.
    """
    ignore_prefixes = [
        "intermediate",
        "_common_metadata",
        "_metadata",
    ]

    catalog_path = get_upath(catalog_path)
    dataset_subdir = catalog_path / "dataset"
    (dataset_path, dataset) = file_io.read_parquet_dataset(
        dataset_subdir,
        ignore_prefixes=ignore_prefixes,
        exclude_invalid_files=True,
    )
    metadata_collector = []
    # Collect the healpix pixels so we can sort before writing.
    healpix_pixels = []
    total_rows = 0

    for single_file in dataset.files:
        relative_path = single_file[len(dataset_path) + 1 :]
        single_metadata = file_io.read_parquet_metadata(dataset_subdir / relative_path)

        # Users must set the file path of each chunk before combining the metadata.
        single_metadata.set_file_path(relative_path)

        if order_by_healpix:
            healpix_pixel = paths.get_healpix_from_path(relative_path)

            healpix_pixels.append(healpix_pixel)
        metadata_collector.append(single_metadata)
        total_rows += single_metadata.num_rows

    ## Write out the two metadata files
    if output_path is None:
        output_path = catalog_path
    if order_by_healpix:
        argsort = get_pixel_argsort(healpix_pixels)
        metadata_collector = np.array(metadata_collector)[argsort]
    catalog_base_dir = get_upath(output_path)
    file_io.make_directory(catalog_base_dir / "dataset", exist_ok=True)
    metadata_file_pointer = paths.get_parquet_metadata_pointer(catalog_base_dir)
    common_metadata_file_pointer = paths.get_common_metadata_pointer(catalog_base_dir)

    file_io.write_parquet_metadata(
        dataset.schema,
        metadata_file_pointer,
        metadata_collector=metadata_collector,
        write_statistics=True,
    )
    file_io.write_parquet_metadata(dataset.schema, common_metadata_file_pointer)
    return total_rows


def read_row_group_fragments(metadata_file: str):
    """Generator for metadata fragment row groups in a parquet metadata file.

    Args:
        metadata_file (str): path to `_metadata` file.
    """
    metadata_file = get_upath(metadata_file)
    if not file_io.is_regular_file(metadata_file):
        metadata_file = paths.get_parquet_metadata_pointer(metadata_file)

    dataset = pds.parquet_dataset(metadata_file, filesystem=metadata_file.fs)

    for frag in dataset.get_fragments():
        yield from frag.row_groups


def _nonemin(value1, value2):
    """Similar to numpy's nanmin, but excludes `None` values.

    NB: If both values are `None`, this will still return `None`"""
    if value1 is None:
        return value2
    if value2 is None:
        return value1
    return min(value1, value2)


def _nonemax(value1, value2):
    """Similar to numpy's nanmax, but excludes `None` values.

    NB: If both values are `None`, this will still return `None`"""
    if value1 is None:
        return value2
    if value2 is None:
        return value1
    return max(value1, value2)


def _pick_columns(
    first_row_group,
    exclude_hats_columns: bool = True,
    exclude_columns: list[str] = None,
    include_columns: list[str] = None,
):
    """Convenience method to find the desired columns and their indexes, given
    some conventional user preferences."""

    if include_columns is None:
        include_columns = []

    if exclude_columns is None:
        exclude_columns = []
    if exclude_hats_columns:
        exclude_columns.extend(["Norder", "Dir", "Npix", "_healpix_29"])

    column_names = [
        first_row_group.column(col).path_in_schema for col in range(0, first_row_group.num_columns)
    ]
    column_names = [name.removesuffix(".list.element") for name in column_names]
    good_column_indexes = [
        index
        for index, name in enumerate(column_names)
        if (len(include_columns) == 0 or name in include_columns)
        and not (len(exclude_columns) > 0 and name in exclude_columns)
    ]
    column_names = [column_names[i] for i in good_column_indexes]

    return good_column_indexes, column_names


def aggregate_column_statistics(
    metadata_file: str | Path | UPath,
    exclude_hats_columns: bool = True,
    exclude_columns: list[str] = None,
    include_columns: list[str] = None,
    include_pixels: list[HealpixPixel] = None,
):
    """Read footer statistics in parquet metadata, and report on global min/max values.

    Args:
        metadata_file (str | Path | UPath): path to `_metadata` file
        exclude_hats_columns (bool): exclude HATS spatial and partitioning fields
            from the statistics. Defaults to True.
        exclude_columns (List[str]): additional columns to exclude from the statistics.
        include_columns (List[str]): if specified, only return statistics for the column
            names provided. Defaults to None, and returns all non-hats columns.
        include_pixels (list[HealpixPixel]): if specified, only return statistics
            for the pixels indicated. Defaults to none, and returns all pixels.
    Returns:
        dataframe with global summary statistics
    """
    total_metadata = file_io.read_parquet_metadata(metadata_file)
    num_row_groups = total_metadata.num_row_groups
    first_row_group = total_metadata.row_group(0)

    good_column_indexes, column_names = _pick_columns(
        first_row_group=first_row_group,
        exclude_hats_columns=exclude_hats_columns,
        exclude_columns=exclude_columns,
        include_columns=include_columns,
    )
    if not good_column_indexes:
        return pd.DataFrame()

    extrema = None

    for row_group_index in range(0, num_row_groups):
        row_group = total_metadata.row_group(row_group_index)
        if include_pixels is not None:
            pixel = paths.get_healpix_from_path(row_group.column(0).file_path)
            if pixel not in include_pixels:
                continue
        row_stats = [
            (
                (None, None, 0, 0)
                if row_group.column(col).statistics is None
                else (
                    row_group.column(col).statistics.min,
                    row_group.column(col).statistics.max,
                    row_group.column(col).statistics.null_count,
                    row_group.column(col).num_values,
                )
            )
            for col in good_column_indexes
        ]
        if extrema is None:
            extrema = row_stats
        ## This is annoying, but avoids extra copies, or none comparison.
        else:
            extrema = [
                (
                    (_nonemin(extrema[col][0], row_stats[col][0])),
                    (_nonemax(extrema[col][1], row_stats[col][1])),
                    extrema[col][2] + row_stats[col][2],
                    extrema[col][3] + row_stats[col][3],
                )
                for col in range(0, len(good_column_indexes))
            ]

    if extrema is None:
        return pd.DataFrame()

    stats_lists = np.array(extrema).T

    frame = pd.DataFrame(
        {
            "column_names": column_names,
            "min_value": stats_lists[0],
            "max_value": stats_lists[1],
            "null_count": stats_lists[2],
            "row_count": stats_lists[3],
        }
    ).set_index("column_names")
    return frame


def per_pixel_statistics(
    metadata_file: str | Path | UPath,
    exclude_hats_columns: bool = True,
    exclude_columns: list[str] = None,
    include_columns: list[str] = None,
    include_stats: list[str] = None,
    multi_index=False,
    include_pixels: list[HealpixPixel] = None,
):
    """Read footer statistics in parquet metadata, and report on statistics about
    each pixel partition.

    Args:
        metadata_file (str | Path | UPath): path to `_metadata` file
        exclude_hats_columns (bool): exclude HATS spatial and partitioning fields
            from the statistics. Defaults to True.
        exclude_columns (List[str]): additional columns to exclude from the statistics.
        include_columns (List[str]): if specified, only return statistics for the column
            names provided. Defaults to None, and returns all non-hats columns.
        include_pixels (list[HealpixPixel]): if specified, only return statistics
            for the pixels indicated. Defaults to none, and returns all pixels.
        include_stats (List[str]): if specified, only return the kinds of values from list
            (min_value, max_value, null_count, row_count). Defaults to None, and returns all values.
        multi_index (bool): should the returned frame be created with a multi-index, first on
            pixel, then on column name? Default is False, and instead indexes on pixel, with
            separate columns per-data-column and stat value combination.
    Returns:
        dataframe with granular per-pixel statistics
    """
    total_metadata = file_io.read_parquet_metadata(metadata_file)
    num_row_groups = total_metadata.num_row_groups
    first_row_group = total_metadata.row_group(0)

    good_column_indexes, column_names = _pick_columns(
        first_row_group=first_row_group,
        exclude_hats_columns=exclude_hats_columns,
        exclude_columns=exclude_columns,
        include_columns=include_columns,
    )
    if not good_column_indexes:
        return pd.DataFrame()

    all_stats = ["min_value", "max_value", "null_count", "row_count"]

    if include_stats is None or len(include_stats) == 0:
        include_stats = all_stats
    else:
        for stat in include_stats:
            if stat not in all_stats:
                raise ValueError(f"include_stats must be from list {all_stats} (found {stat})")

    stat_mask = np.array([ind for ind, stat in enumerate(all_stats) if stat in include_stats])
    pixels = []
    leaf_stats = []

    for row_group_index in range(0, num_row_groups):
        row_group = total_metadata.row_group(row_group_index)
        pixel = paths.get_healpix_from_path(row_group.column(0).file_path)
        if include_pixels is not None and pixel not in include_pixels:
            continue
        row_stats = [
            (
                [None, None, 0, 0]
                if row_group.column(col).statistics is None
                else [
                    row_group.column(col).statistics.min,
                    row_group.column(col).statistics.max,
                    row_group.column(col).statistics.null_count,
                    row_group.column(col).num_values,
                ]
            )
            for col in good_column_indexes
        ]
        row_stats = np.take(row_stats, stat_mask, axis=1)
        pixels.append(pixel)
        leaf_stats.append(row_stats)

    stats_lists = np.array(leaf_stats)
    original_shape = stats_lists.shape

    if multi_index:
        stats_lists = stats_lists.reshape((original_shape[0] * original_shape[1], original_shape[2]))
        frame = pd.DataFrame(
            stats_lists,
            index=pd.MultiIndex.from_product([pixels, column_names], names=["pixel", "column"]),
            columns=include_stats,
        )
    else:
        stats_lists = stats_lists.reshape((original_shape[0], original_shape[1] * original_shape[2]))
        mod_col_names = [[f"{col_name}: {stat}" for stat in include_stats] for col_name in column_names]
        mod_col_names = np.array(mod_col_names).flatten()
        frame = pd.DataFrame(stats_lists, index=pixels, columns=mod_col_names)
    return frame



"""Container class to hold catalog metadata and partition iteration"""

from __future__ import annotations

from hats.catalog.healpix_dataset.healpix_dataset import HealpixDataset
from hats.pixel_math import HealpixPixel
from hats.pixel_tree.negative_tree import compute_negative_tree_pixels


class Catalog(HealpixDataset):
    """A HATS Catalog with data stored in a HEALPix Hive partitioned structure

    Catalogs of this type are partitioned spatially, contain `partition_info` metadata specifying
    the pixels in Catalog, and on disk conform to the parquet partitioning structure
    `Norder=/Dir=/Npix=.parquet`
    """

    def generate_negative_tree_pixels(self) -> list[HealpixPixel]:
        """Get the leaf nodes at each healpix order that have zero catalog data.

        For example, if an example catalog only had data points in pixel 0 at
        order 0, then this method would return order 0's pixels 1 through 11.
        Used for getting full coverage on margin caches.

        Returns:
            List of HealpixPixels representing the 'negative tree' for the catalog.
        """
        return compute_negative_tree_pixels(self.pixel_tree)



from __future__ import annotations

from hats.catalog.healpix_dataset.healpix_dataset import HealpixDataset


class MapCatalog(HealpixDataset):
    """A HATS table to represent non-point-source data in a continuous map."""



from __future__ import annotations

import math

import astropy.units as u
import cdshealpix
import numpy as np
from astropy.coordinates import Latitude, Longitude, SkyCoord

# pylint: disable=missing-function-docstring

## Arithmetic conversions


MAX_HEALPIX_ORDER = 29


def is_order_valid(order: int) -> bool:
    return np.all(0 <= order) and np.all(order <= MAX_HEALPIX_ORDER)


def npix2order(npix: int) -> int:
    if npix <= 0:
        raise ValueError("Invalid value for npix")
    order = int(math.log2(npix / 12)) >> 1
    if not is_order_valid(order) or not 12 * (1 << (2 * order)) == npix:
        raise ValueError("Invalid value for npix")
    return order


def order2nside(order: int) -> int:
    if not is_order_valid(order):
        raise ValueError("Invalid value for order")
    return 1 << order


def order2npix(order: int) -> int:
    if not is_order_valid(order):
        raise ValueError("Invalid value for order")
    return 12 * (1 << (2 * order))


def order2resol(order: int, *, arcmin: bool = False, unit=u.rad) -> float:
    if arcmin:
        unit = u.arcmin
    unit = u.Unit(unit)

    return np.sqrt(order2pixarea(order, unit=unit * unit))


def order2pixarea(order: int, *, degrees: bool = False, unit=u.sr) -> float:
    if degrees:
        unit = "deg**2"
    unit = u.Unit(unit)

    npix = order2npix(order)
    pix_area_rad = 4 * np.pi / npix * u.steradian
    return pix_area_rad.to_value(unit)


def radec2pix(order: int, ra: float, dec: float) -> np.ndarray[np.int64]:
    if not is_order_valid(order):
        raise ValueError("Invalid value for order")

    ra = Longitude(np.asarray(ra, dtype=np.float64), unit="deg")
    dec = Latitude(np.asarray(dec, dtype=np.float64), unit="deg")

    return cdshealpix.lonlat_to_healpix(ra, dec, order).astype(np.int64)


## Coordinate conversion


def ang2vec(ra, dec, **kwargs) -> np.ndarray:
    """Converts ra and dec to cartesian coordinates on the unit sphere"""
    coords = SkyCoord(ra=ra * u.deg, dec=dec * u.deg, **kwargs).cartesian
    return np.array([coords.x.value, coords.y.value, coords.z.value]).T


## Custom functions


def avgsize2mindist(avg_size: np.ndarray) -> np.ndarray:
    """Get the minimum distance between pixels for a given average size

    We don't have the precise geometry of the healpix grid yet,
    so we are using average_size / mininimum_distance = 1.6
    as a rough estimate.

    Parameters
    ----------
    avg_size : np.ndarray of float
        The average size of a healpix pixel

    Returns
    -------
    np.ndarray of float
        The minimum distance between pixels for the given average size
    """
    return avg_size / 1.6


def mindist2avgsize(mindist: np.ndarray) -> np.ndarray:
    """Get the average size for a given minimum distance between pixels

    We don't have the precise geometry of the healpix grid yet,
    so we are using average_size / mininimum_distance = 1.6
    as a rough estimate.

    Parameters
    ----------
    mindist : np.ndarray of float
        The minimum distance between pixels

    Returns
    -------
    np.ndarray of float
        The average size of a healpix pixel for the given minimum distance
        between pixels.
    """
    return mindist * 1.6


def avgsize2order(avg_size_arcmin: np.ndarray) -> np.ndarray:
    """Get the largest order with average healpix size larger than avg_size_arcmin

    Parameters
    ----------
    avg_size_arcmin : np.ndarray of float
        The average size of a healpix pixel in arcminutes

    Returns
    -------
    np.ndarray of int
        The largest healpix order for which the average size is larger than avg_size_arcmin
    """
    # resolution = sqrt(4pi / (12 * 4**order)) => order = log2(sqrt(4pi / 12) / resolution)
    order_float = np.log2(np.sqrt(np.pi / 3) / np.radians(avg_size_arcmin / 60.0))
    return np.array(order_float, dtype=int)


def margin2order(margin_thr_arcmin: np.ndarray) -> np.ndarray:
    """Get the largest order for which distance between pixels is less than margin_thr_arcmin

    We don't have the precise geometry of the healpix grid yet,
    we are using average_size / mininimum_distance = 1.6
    as a rough estimate.

    Parameters
    ----------
    margin_thr_arcmin : np.ndarray of float
        The minimum distance between pixels in arcminutes

    Returns
    -------
    np.ndarray of int
        The largest healpix order for which the distance between pixels is less than margin_thr_arcmin
    """
    avg_size_arcmin = mindist2avgsize(margin_thr_arcmin)
    return avgsize2order(avg_size_arcmin)


def order2mindist(order: np.ndarray | int) -> np.ndarray | float:
    """Get the estimated minimum distance between pixels at a given order.

    We don't have the precise geometry of the healpix grid yet,
    we are using average_size / mininimum_distance = 1.6
    as a rough estimate.

    Parameters
    ----------
    order : np.ndarray of int or a single scalar int
        The healpix order

    Returns
    -------
    np.ndarray of float or a single scalar float
        The minimum distance between pixels in arcminutes
    """
    pixel_avgsize = order2resol(order, arcmin=True)
    return avgsize2mindist(pixel_avgsize)



import re
from functools import reduce
from pathlib import Path
from typing import Iterable, Optional

from jproperties import Properties
from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator, model_validator
from typing_extensions import Self
from upath import UPath

from hats.io import file_io

# All additional properties in the HATS recommendation.
EXTRA_ALLOWED_FIELDS = [
    "addendum_did",
    "bib_reference",
    "bib_reference_url",
    "creator_did",
    "data_ucd",
    "hats_builder",
    "hats_coordinate_epoch",
    "hats_copyright",
    "hats_creation_date",
    "hats_creator",
    "hats_estsize",
    "hats_progenitor_url",
    "hats_release_date",
    "hats_service_url",
    "hats_status",
    "hats_version",
    "moc_sky_fraction",
    "obs_ack",
    "obs_copyright",
    "obs_copyright_url",
    "obs_description",
    "obs_regime",
    "obs_title",
    "prov_progenitor",
    "publisher_id",
    "t_max",
    "t_min",
]


class CollectionProperties(BaseModel):
    """Container class for catalog metadata"""

    name: str = Field(alias="obs_collection")

    hats_primary_table_url: Optional[str] = Field(default=None, alias="hats_primary_table_url")
    """Reference to object catalog. Relevant for nested, margin, association, and index."""

    all_margins: Optional[list[str]] = Field(default=None)
    default_margin: Optional[str] = Field(default=None)

    all_indexes: Optional[dict[str, str]] = Field(default=None)
    default_index: Optional[str] = Field(default=None)

    ## Allow any extra keyword args to be stored on the properties object.
    model_config = ConfigDict(extra="allow", populate_by_name=True, use_enum_values=True)

    @field_validator("all_margins", mode="before")
    @classmethod
    def space_delimited_list(cls, str_value: str) -> list[str]:
        """Convert a space-delimited list string into a python list of strings."""
        if not str_value or not isinstance(str_value, str):
            ## Convert empty strings and empty lists to None
            return None
        # Split on a few kinds of delimiters (just to be safe), and remove duplicates
        return list(filter(None, re.split(";| |,|\n", str_value)))

    @field_validator("all_indexes", mode="before")
    @classmethod
    def index_tuples(cls, str_value: str) -> dict[str, str]:
        """Convert a space-delimited list string into a python list of strings."""
        if not str_value or not isinstance(str_value, str):
            return None
        # Split on a few kinds of delimiters (just to be safe), and remove duplicates
        str_values = list(filter(None, re.split(";| |,|\n", str_value)))
        ## Convert empty strings and empty lists to None
        if len(str_values) % 2 != 0:
            raise ValueError("Collection all_indexes map should contain pairs of field and index name")
        all_index_dict = {}
        for index_start in range(0, len(str_values), 2):
            key = str_values[index_start]
            value = str_values[index_start + 1]
            all_index_dict[key] = value
        return all_index_dict

    @field_serializer("all_margins")
    def serialize_list_as_space_delimited_list(self, str_list: Iterable[str]) -> str:
        """Convert a python list of strings into a space-delimited string."""
        if str_list is None or len(str_list) == 0:
            return ""
        return " ".join(str_list)

    @field_serializer("all_indexes")
    def serialize_dict_as_space_delimited_list(self, str_dict: dict[str, str]) -> str:
        """Convert a python list of strings into a space-delimited string."""
        if str_dict is None or len(str_dict) == 0:
            return ""
        str_list = list(reduce(lambda x, y: x + y, str_dict.items()))
        return " ".join(str_list)

    @model_validator(mode="after")
    def check_allowed_and_required(self) -> Self:
        """Check that type-specific fields are appropriate, and required fields are set."""
        # Check against all known properties - catches typos.
        non_allowed = set(self.__pydantic_extra__.keys()) - set(EXTRA_ALLOWED_FIELDS)
        if len(non_allowed) > 0:
            raise ValueError(f"Unexpected extra property: {non_allowed}")
        return self

    def explicit_dict(self):
        """Create a dict, based on fields that have been explicitly set, and are not "extra" keys."""
        explicit = self.model_dump(by_alias=False, exclude_none=True)
        extra_keys = self.__pydantic_extra__.keys()
        return {key: val for key, val in explicit.items() if key not in extra_keys}

    def __str__(self):
        """Friendly string representation based on named fields."""
        parameters = self.explicit_dict()
        formatted_string = ""
        for name, value in parameters.items():
            formatted_string += f"  {name} {value}\n"
        return formatted_string

    @classmethod
    def read_from_dir(cls, catalog_dir: str | Path | UPath) -> Self:
        """Read field values from a java-style properties file."""
        file_path = file_io.get_upath(catalog_dir) / "collection.properties"
        if not file_io.does_file_or_directory_exist(file_path):
            raise FileNotFoundError(f"No properties file found where expected: {str(file_path)}")
        p = Properties()
        with file_path.open("rb") as f:
            p.load(f, "utf-8")
        return cls(**p.properties)

    def to_properties_file(self, catalog_dir: str | Path | UPath) -> Self:
        """Write fields to a java-style properties file."""
        # pylint: disable=protected-access
        parameters = self.model_dump(by_alias=True, exclude_none=True)
        properties = Properties()
        properties.properties = parameters
        properties._key_order = parameters.keys()
        file_path = file_io.get_upath(catalog_dir) / "collection.properties"
        with file_path.open("wb") as _file:
            properties.store(_file, encoding="utf-8", initial_comments="HATS Collection", timestamp=False)



from __future__ import annotations

import warnings
from pathlib import Path

import astropy.units as u
import numpy as np
import pandas as pd
import pyarrow as pa
from astropy.coordinates import SkyCoord
from mocpy import MOC
from typing_extensions import Self
from upath import UPath

from hats.catalog.dataset import Dataset
from hats.catalog.dataset.table_properties import TableProperties
from hats.catalog.partition_info import PartitionInfo
from hats.inspection import plot_pixels
from hats.inspection.visualize_catalog import plot_moc
from hats.io.parquet_metadata import aggregate_column_statistics, per_pixel_statistics
from hats.pixel_math import HealpixPixel
from hats.pixel_math.box_filter import generate_box_moc, wrap_ra_angles
from hats.pixel_math.validators import (
    validate_box,
    validate_declination_values,
    validate_polygon,
    validate_radius,
)
from hats.pixel_tree import PixelAlignment, PixelAlignmentType
from hats.pixel_tree.moc_filter import filter_by_moc
from hats.pixel_tree.pixel_alignment import align_with_mocs
from hats.pixel_tree.pixel_tree import PixelTree


class HealpixDataset(Dataset):
    """A HATS dataset partitioned with a HEALPix partitioning structure.

    Catalogs of this type are partitioned based on the ra and dec of the points with each partition
    containing points within a given HEALPix pixel. The files are in the form::

        Norder=/Dir=/Npix=.parquet
    """

    def __init__(
        self,
        catalog_info: TableProperties,
        pixels: PartitionInfo | PixelTree | list[HealpixPixel],
        catalog_path: str | Path | UPath | None = None,
        moc: MOC | None = None,
        schema: pa.Schema | None = None,
    ) -> None:
        """Initializes a Catalog

        Args:
            catalog_info: TableProperties object with catalog metadata
            pixels: Specifies the pixels contained in the catalog. Can be either a
                list of HealpixPixel, `PartitionInfo object`, or a `PixelTree` object
            catalog_path: If the catalog is stored on disk, specify the location of the catalog
                Does not load the catalog from this path, only store as metadata
            moc (mocpy.MOC): MOC object representing the coverage of the catalog
            schema (pa.Schema): The pyarrow schema for the catalog
        """
        super().__init__(catalog_info, catalog_path=catalog_path, schema=schema)
        self.partition_info = self._get_partition_info_from_pixels(pixels)
        self.pixel_tree = self._get_pixel_tree_from_pixels(pixels)
        self.moc = moc

    def get_healpix_pixels(self) -> list[HealpixPixel]:
        """Get healpix pixel objects for all pixels contained in the catalog.

        Returns:
            List of HealpixPixel
        """
        return self.partition_info.get_healpix_pixels()

    @staticmethod
    def _get_partition_info_from_pixels(
        pixels: PartitionInfo | PixelTree | list[HealpixPixel],
    ) -> PartitionInfo:
        if isinstance(pixels, PartitionInfo):
            return pixels
        if isinstance(pixels, PixelTree):
            return PartitionInfo.from_healpix(pixels.get_healpix_pixels())
        if pd.api.types.is_list_like(pixels):
            return PartitionInfo.from_healpix(pixels)
        raise TypeError("Pixels must be of type PartitionInfo, PixelTree, or List[HealpixPixel]")

    @staticmethod
    def _get_pixel_tree_from_pixels(pixels: PartitionInfo | PixelTree | list[HealpixPixel]) -> PixelTree:
        if isinstance(pixels, PartitionInfo):
            return PixelTree.from_healpix(pixels.get_healpix_pixels())
        if isinstance(pixels, PixelTree):
            return pixels
        if pd.api.types.is_list_like(pixels):
            return PixelTree.from_healpix(pixels)
        raise TypeError("Pixels must be of type PartitionInfo, PixelTree, or List[HealpixPixel]")

    def __len__(self):
        """The number of rows in the catalog.

        Returns:
            The number of rows in the catalog, as specified in its metadata.
            This value is undetermined when the catalog is modified, and
            therefore an error is raised.
        """
        if self.catalog_info.total_rows == 0:
            raise ValueError("The number of rows is undetermined because the catalog was modified.")
        return self.catalog_info.total_rows

    def get_max_coverage_order(self) -> int:
        """Gets the maximum HEALPix order for which the coverage of the catalog is known from the pixel
        tree and moc if it exists"""
        if len(self.pixel_tree) == 0:
            raise ValueError("Cannot get max_order of empty catalog")
        max_order = (
            max(self.moc.max_order, self.pixel_tree.get_max_depth())
            if self.moc is not None
            else self.pixel_tree.get_max_depth()
        )
        return max_order

    def filter_from_pixel_list(self, pixels: list[HealpixPixel]) -> Self:
        """Filter the pixels in the catalog to only include any that overlap with the requested pixels.

        Args:
            pixels (List[HealpixPixels]): the pixels to include

        Returns:
            A new catalog with only the pixels that overlap with the given pixels. Note that we reset the
            total_rows to None, as updating would require a scan over the new pixel sizes.
        """
        orders = np.array([p.order for p in pixels])
        pixel_inds = np.array([p.pixel for p in pixels])
        max_order = np.max(orders) if len(orders) > 0 else 0
        moc = MOC.from_healpix_cells(ipix=pixel_inds, depth=orders, max_depth=max_order)
        return self.filter_by_moc(moc)

    def filter_by_cone(self, ra: float, dec: float, radius_arcsec: float) -> Self:
        """Filter the pixels in the catalog to only include the pixels that overlap with a cone

        Args:
            ra (float): Right ascension of the center of the cone, in degrees
            dec (float): Declination of the center of the cone, in degrees
            radius_arcsec (float): Radius of the cone, in arcseconds

        Returns:
            A new catalog with only the pixels that overlap with the specified cone
        """
        validate_radius(radius_arcsec)
        validate_declination_values(dec)
        cone_moc = MOC.from_cone(
            lon=ra * u.deg,
            lat=dec * u.deg,
            radius=radius_arcsec * u.arcsec,
            max_depth=self.get_max_coverage_order(),
        )
        return self.filter_by_moc(cone_moc)

    def filter_by_box(self, ra: tuple[float, float], dec: tuple[float, float]) -> Self:
        """Filter the pixels in the catalog to only include the pixels that overlap with a
        zone, defined by right ascension and declination ranges. The right ascension edges follow
        great arc circles and the declination edges follow small arc circles.

        Args:
            ra (Tuple[float, float]): Right ascension range, in degrees
            dec (Tuple[float, float]): Declination range, in degrees

        Returns:
            A new catalog with only the pixels that overlap with the specified region
        """
        ra = tuple(wrap_ra_angles(ra)) if ra else None
        validate_box(ra, dec)
        box_moc = generate_box_moc(ra, dec, self.get_max_coverage_order())
        return self.filter_by_moc(box_moc)

    def filter_by_polygon(self, vertices: list[tuple[float, float]]) -> Self:
        """Filter the pixels in the catalog to only include the pixels that overlap
        with a polygonal sky region.

        Args:
            vertices (list[tuple[float,float]]): The list of vertice coordinates for
                the polygon, (ra, dec), in degrees.

        Returns:
            A new catalog with only the pixels that overlap with the specified polygon.
        """
        validate_polygon(vertices)
        polygon_moc = MOC.from_polygon_skycoord(
            SkyCoord(vertices, unit="deg"), max_depth=self.get_max_coverage_order()
        )
        return self.filter_by_moc(polygon_moc)

    def filter_by_moc(self, moc: MOC) -> Self:
        """Filter the pixels in the catalog to only include the pixels that overlap with the moc provided.

        Args:
            moc (mocpy.MOC): the moc to filter by

        Returns:
            A new catalog with only the pixels that overlap with the moc. Note that we reset the total_rows
        to 0, as updating would require a scan over the new pixel sizes."""
        filtered_tree = filter_by_moc(self.pixel_tree, moc)
        filtered_moc = self.moc.intersection(moc) if self.moc is not None else None
        filtered_catalog_info = self.catalog_info.copy_and_update(total_rows=0)
        return self.__class__(
            filtered_catalog_info,
            pixels=filtered_tree,
            catalog_path=self.catalog_path,
            moc=filtered_moc,
            schema=self.schema,
        )

    def align(
        self, other_cat: Self, alignment_type: PixelAlignmentType = PixelAlignmentType.INNER
    ) -> PixelAlignment:
        """Performs an alignment to another catalog, using the pixel tree and mocs if available

        An alignment compares the pixel structures of the two catalogs, checking which pixels overlap.
        The alignment includes the mapping of all pairs of pixels in each tree that overlap with each other,
        and the aligned tree which consists of the overlapping pixels in the two input catalogs, using the
        higher order pixels where there is overlap with differing orders.

        For more information, see this document:
        https://docs.google.com/document/d/1gqb8qb3HiEhLGNav55LKKFlNjuusBIsDW7FdTkc5mJU/edit?usp=sharing

        Args:
            other_cat (Catalog): The catalog to align to
            alignment_type (PixelAlignmentType): The type of alignment describing how to handle nodes which
            exist in one tree but not the other. Mirrors the 'how' argument of a pandas/sql join. Options are:

                - "inner" - only use pixels that appear in both catalogs
                - "left" - use all pixels that appear in the left catalog and any overlapping from the right
                - "right" - use all pixels that appear in the right catalog and any overlapping from the left
                - "outer" - use all pixels from both catalogs

        Returns (PixelAlignment):
            A `PixelAlignment` object with the alignment from the two catalogs
        """
        return align_with_mocs(
            self.pixel_tree, other_cat.pixel_tree, self.moc, other_cat.moc, alignment_type=alignment_type
        )

    def plot_pixels(self, **kwargs):
        """Create a visual map of the pixel density of the catalog.

        Args:
            kwargs: Additional args to pass to `hats.inspection.visualize_catalog.plot_healpix_map`
        """
        return plot_pixels(self, **kwargs)

    def plot_moc(self, **kwargs):
        """Create a visual map of the coverage of the catalog.

        Args:
            kwargs: Additional args to pass to `hats.inspection.visualize_catalog.plot_moc`
        """
        default_title = f"Coverage MOC of {self.catalog_name}"
        plot_args = {"title": default_title}
        plot_args.update(kwargs)
        return plot_moc(self.moc, **plot_args)

    def aggregate_column_statistics(
        self,
        exclude_hats_columns: bool = True,
        exclude_columns: list[str] | None = None,
        include_columns: list[str] | None = None,
        include_pixels: list[HealpixPixel] | None = None,
    ):
        """Read footer statistics in parquet metadata, and report on global min/max values.

        Args:
            exclude_hats_columns (bool): exclude HATS spatial and partitioning fields
                from the statistics. Defaults to True.
            exclude_columns (List[str]): additional columns to exclude from the statistics.
            include_columns (List[str]): if specified, only return statistics for the column
                names provided. Defaults to None, and returns all non-hats columns.
        """
        if not self.on_disk:
            warnings.warn("Calling aggregate_column_statistics on an in-memory catalog. No results.")
            return pd.DataFrame()

        if include_pixels is None:
            include_pixels = self.get_healpix_pixels()
        return aggregate_column_statistics(
            self.catalog_base_dir / "dataset" / "_metadata",
            exclude_hats_columns=exclude_hats_columns,
            exclude_columns=exclude_columns,
            include_columns=include_columns,
            include_pixels=include_pixels,
        )

    def per_pixel_statistics(
        self,
        exclude_hats_columns: bool = True,
        exclude_columns: list[str] | None = None,
        include_columns: list[str] | None = None,
        include_stats: list[str] | None = None,
        multi_index=False,
        include_pixels: list[HealpixPixel] | None = None,
    ):
        """Read footer statistics in parquet metadata, and report on statistics about
        each pixel partition.

        Args:
            exclude_hats_columns (bool): exclude HATS spatial and partitioning fields
                from the statistics. Defaults to True.
            exclude_columns (List[str]): additional columns to exclude from the statistics.
            include_columns (List[str]): if specified, only return statistics for the column
                names provided. Defaults to None, and returns all non-hats columns.
            include_pixels (list[HealpixPixel]): if specified, only return statistics
                for the pixels indicated. Defaults to none, and returns all pixels.
            include_stats (List[str]): if specified, only return the kinds of values from list
                (min_value, max_value, null_count, row_count). Defaults to None, and returns all values.
            multi_index (bool): should the returned frame be created with a multi-index, first on
                pixel, then on column name?
        """
        if not self.on_disk:
            warnings.warn("Calling per_pixel_statistics on an in-memory catalog. No results.")
            return pd.DataFrame()

        if include_pixels is None:
            include_pixels = self.get_healpix_pixels()
        return per_pixel_statistics(
            self.catalog_base_dir / "dataset" / "_metadata",
            exclude_hats_columns=exclude_hats_columns,
            exclude_columns=exclude_columns,
            include_columns=include_columns,
            include_stats=include_stats,
            multi_index=multi_index,
            include_pixels=include_pixels,
        )
